from datetime import datetime
from typing import Callable

from combustion_ble.uart import (
    ReadOverTemperatureResponse,
    SetColorResponse,
    SetIDResponse,
    SetPredictionResponse,
)
from combustion_ble.uart.meatnet import NodeSetPredictionResponse

SuccessHandler = Callable[[bool], None]
ReadOverTemperatureHandler = Callable[[bool, bool], None]


# Structs to store when BLE message was sent and the completion handler for message
class MessageSentHandler:
    def __init__(
        self,
        time_sent: datetime,
        success_handler: SuccessHandler | None,
        read_over_temperature_completion_handler: ReadOverTemperatureHandler | None,
    ) -> None:
        self.time_sent = time_sent
        self.success_handler = success_handler
        self.read_over_temperature_completion_handler = read_over_temperature_completion_handler


class MessageHandlers:
    MESSAGE_TIMEOUT_SECONDS = 3

    def __init__(self):
        self.set_id_completion_handlers: dict[str, MessageSentHandler] = {}
        self.set_color_completion_handlers: dict[str, MessageSentHandler] = {}
        self.set_prediction_completion_handlers: dict[str, MessageSentHandler] = {}
        self.read_over_temperature_completion_handlers: dict[str, MessageSentHandler] = {}
        self.set_node_prediction_completion_handlers: dict[str, MessageSentHandler] = {}

    def check_for_timeout(self):
        current_time = datetime.now()
        self._check_for_message_timeout(self.set_id_completion_handlers, current_time)
        self._check_for_message_timeout(self.set_color_completion_handlers, current_time)
        self._check_for_message_timeout(self.set_prediction_completion_handlers, current_time)
        self._check_for_message_timeout(
            self.read_over_temperature_completion_handlers, current_time
        )
        self._check_for_message_timeout(self.set_node_prediction_completion_handlers, current_time)

    def _check_for_message_timeout(
        self, handlers: dict[str, MessageSentHandler], current_time: datetime
    ):
        keys_to_remove = []
        for key, value in handlers.items():
            if (current_time - value.time_sent).total_seconds() > self.MESSAGE_TIMEOUT_SECONDS:
                if value.success_handler:
                    value.success_handler(False)
                if value.read_over_temperature_completion_handler:
                    value.read_over_temperature_completion_handler(False, False)
                keys_to_remove.append(key)

        for key in keys_to_remove:
            del handlers[key]

    def clear_handlers_for_device(self, device_identifier: str):
        if device_identifier in self.set_color_completion_handlers:
            del self.set_color_completion_handlers[device_identifier]

        if device_identifier in self.set_id_completion_handlers:
            del self.set_id_completion_handlers[device_identifier]

        if device_identifier in self.set_prediction_completion_handlers:
            del self.set_prediction_completion_handlers[device_identifier]

        if device_identifier in self.read_over_temperature_completion_handlers:
            del self.read_over_temperature_completion_handlers[device_identifier]

        if device_identifier in self.set_node_prediction_completion_handlers:
            del self.set_node_prediction_completion_handlers[device_identifier]

    def add_set_id_completion_handler(
        self, device_identifier: str, completion_handler: SuccessHandler
    ):
        self.set_id_completion_handlers[device_identifier] = MessageSentHandler(
            datetime.now(), completion_handler, None
        )

    def call_set_id_completion_handler(self, identifier: str, response: SetIDResponse):
        handler = self.set_id_completion_handlers.get(identifier)
        if handler and handler.success_handler:
            handler.success_handler(response.success)
        self.set_id_completion_handlers.pop(identifier, None)

    def add_set_color_completion_handler(
        self, device_identifier: str, completion_handler: SuccessHandler
    ):
        self.set_color_completion_handlers[device_identifier] = MessageSentHandler(
            datetime.now(), completion_handler, None
        )

    def call_set_color_completion_handler(self, identifier: str, response: SetColorResponse):
        handler = self.set_color_completion_handlers.get(identifier)
        if handler and handler.success_handler:
            handler.success_handler(response.success)
        self.set_color_completion_handlers.pop(identifier, None)

    def add_set_prediction_completion_handler(
        self, device_identifier: str, completion_handler: SuccessHandler
    ):
        self.set_prediction_completion_handlers[device_identifier] = MessageSentHandler(
            datetime.now(), completion_handler, None
        )

    def call_set_prediction_completion_handler(
        self, identifier: str, response: SetPredictionResponse
    ):
        handler = self.set_prediction_completion_handlers.get(identifier)
        if handler and handler.success_handler:
            handler.success_handler(response.success)
        self.set_prediction_completion_handlers.pop(identifier, None)

    def add_read_over_temperature_completion_handler(
        self, device_identifier: str, completion_handler: ReadOverTemperatureHandler
    ):
        self.read_over_temperature_completion_handlers[device_identifier] = MessageSentHandler(
            datetime.now(), None, completion_handler
        )

    def call_read_over_temperature_completion_handler(
        self, identifier: str, response: ReadOverTemperatureResponse
    ):
        handler = self.read_over_temperature_completion_handlers.get(identifier)
        if handler and handler.read_over_temperature_completion_handler:
            handler.read_over_temperature_completion_handler(response.success, response.flag_set)
        self.read_over_temperature_completion_handlers.pop(identifier, None)

    def add_node_set_prediction_completion_handler(
        self, device_identifier: str, completion_handler: SuccessHandler
    ):
        self.set_node_prediction_completion_handlers[device_identifier] = MessageSentHandler(
            datetime.now(), completion_handler, None
        )

    def call_node_set_prediction_completion_handler(
        self, identifier: str, response: NodeSetPredictionResponse
    ):
        handler = self.set_node_prediction_completion_handlers.get(identifier)
        if handler and handler.success_handler:
            handler.success_handler(response.success)
        self.set_node_prediction_completion_handlers.pop(identifier, None)



from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.uart.meatnet.node_request import NodeRequest


class NodeReadModelInfoRequest(NodeRequest):
    def __init__(self, serial_number: int):
        # Packing the serial number into bytes
        serial_number_bytes = serial_number.to_bytes(length=4, byteorder="little")

        # Initialize the base class with the payload and type
        super().__init__(
            outgoing_payload=serial_number_bytes,
            message_type=NodeMessageType.PROBE_MODEL_INFORMATION,
        )



"""Bluetooth Advertising Data."""

from enum import Enum
from typing import NamedTuple, Optional

from combustion_ble.ble_data.battery_status_virtual_sensors import (
    BatteryStatusVirtualSensors,
)
from combustion_ble.ble_data.hop_count import HopCount
from combustion_ble.ble_data.mode_id import ModeId
from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures


class CombustionProductType(Enum):
    """Enumeration representing the types of combustion products."""

    UNKNOWN = 0x00
    PROBE = 0x01
    MEAT_NET_NODE = 0x02


class AdvertisingData(NamedTuple):
    """Bluetooth Advertising Data."""

    type: CombustionProductType
    serial_number: int
    temperatures: ProbeTemperatures
    mode_id: ModeId
    battery_status_virtual_sensors: BatteryStatusVirtualSensors
    hop_count: HopCount
    bit_string: str

    @staticmethod
    def from_data(data: bytes) -> Optional["AdvertisingData"]:
        """Create instance from raw advertising data."""
        from bitstring import Bits

        if data is None or len(data) < 20:
            return None

        # Vendor ID
        vendor_id = int.from_bytes(data[0:2], byteorder="big")
        if vendor_id != 0x09C7:
            return None

        # Product type
        type_byte = data[2]
        product_type = CombustionProductType(type_byte)

        # Serial number
        serial_number = int.from_bytes(data[3:7], byteorder="little")

        # Temperatures
        temperatures = ProbeTemperatures.from_raw_data(data[7:20])

        # ModeId
        mode_id = ModeId.from_byte(data[20]) if len(data) >= 21 else ModeId.default_values()

        # Battery Status and Virtual Sensors
        battery_status_virtual_sensors = (
            BatteryStatusVirtualSensors.from_byte(data[21])
            if len(data) >= 22
            else BatteryStatusVirtualSensors.default_values()
        )

        # Hop Count
        hop_count = (
            HopCount.from_network_info_byte(data[22])
            if len(data) >= 23
            else HopCount.default_values()
        )

        # Bit String
        bit_string = Bits(data).bin

        return AdvertisingData(
            type=product_type,
            serial_number=serial_number,
            temperatures=temperatures,
            mode_id=mode_id,
            battery_status_virtual_sensors=battery_status_virtual_sensors,
            hop_count=hop_count,
            bit_string=bit_string,
        )

    @staticmethod
    def from_bleak_data(data: bytes) -> Optional["AdvertisingData"]:
        """Create instance from raw Bleak advertising data."""

        vendor_id = 0x09C7.to_bytes(2, "big")
        return AdvertisingData.from_data(vendor_id + data)



"""Hop Count."""

from enum import Enum


class HopCount(Enum):
    """Hop Count."""

    HOP1 = 0x00
    HOP2 = 0x01
    HOP3 = 0x02
    HOP4 = 0x03

    HOP_COUNT_MASK = 0x3
    HOP_COUNT_SHIFT = 6

    @staticmethod
    def from_network_info_byte(network_info_byte):
        """Generate hop count from network info byte."""
        raw_hop_count = (
            network_info_byte >> HopCount.HOP_COUNT_SHIFT.value
        ) & HopCount.HOP_COUNT_MASK.value
        return (
            HopCount(raw_hop_count)
            if raw_hop_count in HopCount._value2member_map_
            else HopCount.HOP1
        )

    @staticmethod
    def default_values():
        """Generate default values."""
        return HopCount.HOP1



"""Combustion Bluetooth Parsing."""

from .advertising_data import AdvertisingData, CombustionProductType
from .battery_status_virtual_sensors import BatteryStatus
from .mode_id import ProbeColor, ProbeID, ProbeMode
from .prediction_status import (
    PredictionMode,
    PredictionState,
    PredictionStatus,
    PredictionType,
)
from .probe_temperatures import ProbeTemperatures

__all__ = [
    "AdvertisingData",
    "BatteryStatus",
    "CombustionProductType",
    "ProbeColor",
    "ProbeID",
    "ProbeMode",
    "PredictionMode",
    "PredictionState",
    "PredictionStatus",
    "PredictionType",
    "ProbeTemperatures",
]



import struct

from combustion_ble.uart.meatnet.node_response import NodeResponse


class NodeReadHardwareRevisionResponse(NodeResponse):
    MINIMUM_PAYLOAD_LENGTH = 20
    SERIAL_RANGE = slice(NodeResponse.HEADER_LENGTH, NodeResponse.HEADER_LENGTH + 4)
    HW_REVISION_RANGE = slice(NodeResponse.HEADER_LENGTH + 4, NodeResponse.HEADER_LENGTH + 20)

    def __init__(self, data, success, request_id, response_id, payload_length):
        # Extracting the serial number
        serial_raw = data[self.SERIAL_RANGE]
        self.probe_serial_number = struct.unpack("<I", serial_raw)[0]

        # Extracting the hardware revision
        hw_revision_raw = data[self.HW_REVISION_RANGE]
        self.hw_revision = hw_revision_raw.decode("utf-8").rstrip("\x00")

        super().__init__(success, request_id, response_id, payload_length)

    @classmethod
    def from_raw(cls, data, success, request_id, response_id, payload_length):
        if payload_length < cls.MINIMUM_PAYLOAD_LENGTH:
            return None
        return cls(data, success, request_id, response_id, payload_length)



"""Session Info"""

from combustion_ble.uart.message_type import MessageType
from combustion_ble.uart.request import Request
from combustion_ble.uart.response import Response


class SessionInformation:
    def __init__(self, session_id: int, sample_period: int) -> None:
        self.session_id = session_id
        self.sample_period = sample_period


class SessionInfoRequest(Request):
    def __init__(self):
        super().__init__(payload=b"", message_type=MessageType.SESSION_INFO)


class SessionInfoResponse(Response):
    PAYLOAD_LENGTH = 6

    def __init__(self, data: bytes, success: bool, payload_length: int):
        sequence_byte_index = Response.HEADER_LENGTH
        session_id_raw = data[sequence_byte_index : sequence_byte_index + 4]
        session_id = int.from_bytes(session_id_raw, "little")

        sample_period_raw = data[sequence_byte_index + 4 : sequence_byte_index + 6]
        sample_period = int.from_bytes(sample_period_raw, "little")

        self.info = SessionInformation(session_id, sample_period)
        super().__init__(success, payload_length)

    @staticmethod
    def from_raw(data: bytes, success: bool, payload_length: int):
        if payload_length < SessionInfoResponse.PAYLOAD_LENGTH:
            return None

        return SessionInfoResponse(data, success, payload_length)



import struct

from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.uart.meatnet.node_request import NodeRequest


class NodeReadFirmwareRevisionRequest(NodeRequest):
    def __init__(self, serial_number):
        # Packing the serial number into bytes
        serial_number_bytes = struct.pack("<I", serial_number)

        # Initialize the base class with the payload and type
        super().__init__(
            outgoing_payload=serial_number_bytes,
            message_type=NodeMessageType.PROBE_FIRMWARE_REVISION,
        )



"""Prediction Mode."""

from enum import Enum


class PredictionMode(Enum):
    """Prediction Mode."""

    NONE = 0x00
    TIME_TO_REMOVAL = 0x01
    REMOVAL_AND_RESTING = 0x02
    RESERVED = 0x03

    MASK = 0x3

    def to_string(self):
        """Human-readable representation of prediction mode."""
        if self == PredictionMode.NONE:
            return "None"
        elif self == PredictionMode.TIME_TO_REMOVAL:
            return "Time to Removal"
        elif self == PredictionMode.REMOVAL_AND_RESTING:
            return "Remove and Resting"
        elif self == PredictionMode.RESERVED:
            return "Reserved"



class CombustionError(Exception):
    """Base Combustion Error."""


class CharacteristicMissingError(CombustionError):
    """Raised when a characteristic is missing."""


class DFUNotImplementedError(NotImplementedError, CombustionError):
    """Raised when a DFU Operation is attempted."""

    def __init__(self) -> None:
        super().__init__("DFU Operations are not supported by this SDK")



from combustion_ble.uart.message_type import MessageType
from combustion_ble.uart.request import Request


class LogRequest(Request):
    def __init__(self, min_sequence: int, max_sequence: int):
        # Packing the min and max sequence numbers into binary format
        min_payload = min_sequence.to_bytes(length=4, byteorder="little")
        max_payload = max_sequence.to_bytes(length=4, byteorder="little")

        # Combining the payloads
        combined_payload = min_payload + max_payload

        # Calling the superclass initializer with the combined payload and a type
        super().__init__(payload=combined_payload, message_type=MessageType.LOG)



from typing import TYPE_CHECKING

from combustion_ble.ble_data.advertising_data import AdvertisingData
from combustion_ble.devices.device import Device
from combustion_ble.dfu_manager import DFUDeviceType

if TYPE_CHECKING:
    from ..device_manager import DeviceManager
    from .probe import Probe


class MeatNetNode(Device):
    def __init__(
        self,
        advertising: AdvertisingData,
        device_manager: "DeviceManager",
        is_connectable: bool,
        rssi: int,
        identifier: str,
    ):
        super().__init__(
            unique_identifier=str(identifier),
            device_manager=device_manager,
            ble_identifier=identifier,
            rssi=rssi,
        )
        self.serial_number_string = None
        self.probes: dict[int, Probe] = {}
        self.dfu_type = DFUDeviceType.UNKNOWN
        self.update_with_advertising(advertising, is_connectable, rssi)

    def update_with_advertising(
        self, advertising: AdvertisingData, is_connectable: bool, rssi: int
    ):
        self._rssi.update(rssi)
        self.is_connectable = is_connectable

    def update_networked_probe(self, probe: "Probe"):
        if probe is not None:
            self.probes[probe.serial_number] = probe

    def has_connection_to_probe(self, serial_number: int):
        return serial_number in self.probes or str(serial_number) in self.probes  # todo ... yucky

    def update_with_model_info(self, model_info: str):
        super().update_with_model_info(model_info)
        if "Timer" in model_info:
            self.dfu_type = "display"
        elif "Charger" in model_info:
            self.dfu_type = "charger"

    def __str__(self):
        return f"MeatNetNode: {self.unique_identifier}"



from enum import Enum

from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures


class VirtualCoreSensor(Enum):
    """Virtual Core Sensor."""

    T1 = 0x00
    T2 = 0x01
    T3 = 0x02
    T4 = 0x03
    T5 = 0x04
    T6 = 0x05

    MASK = 0x7

    def sensor_number(self):
        """Display value (1-based) for this sensor."""
        return int(self.value) + 1

    def temperature_from(self, temperatures: ProbeTemperatures):
        """Get temperature for virtual sensor."""
        return temperatures.values[int(self.value)]


class VirtualSurfaceSensor(Enum):
    """Virtual Surface Sensor."""

    T4 = 0x00
    T5 = 0x01
    T6 = 0x02
    T7 = 0x03

    MASK = 0x3

    def sensor_number(self):
        """Display value (1-based) for this sensor."""
        return int(self.value) + 4

    def temperature_from(self, temperatures: ProbeTemperatures):
        """Get temperature for virtual sensor."""
        surface_sensor_number = int(self.value) + 3
        return temperatures.values[surface_sensor_number]


class VirtualAmbientSensor(Enum):
    """Virtual Ambient Sensor."""

    T5 = 0x00
    T6 = 0x01
    T7 = 0x02
    T8 = 0x03

    MASK = 0x3

    def sensor_number(self):
        """Display value (1-based) for this sensor."""
        return int(self.value) + 5

    def temperature_from(self, temperatures: ProbeTemperatures):
        """Get temperature for virtual sensor."""
        ambient_sensor_number = int(self.value) + 4
        return temperatures.values[ambient_sensor_number]


class VirtualSensors:
    """Collection of all virtual sensors."""

    def __init__(
        self,
        virtual_core: VirtualCoreSensor,
        virtual_surface: VirtualSurfaceSensor,
        virtual_ambient: VirtualAmbientSensor,
    ):
        """Initialize."""
        self.virtual_core = virtual_core
        self.virtual_surface = virtual_surface
        self.virtual_ambient = virtual_ambient

    @staticmethod
    def from_byte(byte):
        """Create instances from byte."""
        raw_virtual_core = byte & VirtualCoreSensor.MASK.value
        try:
            virtual_core = VirtualCoreSensor(raw_virtual_core)
        except ValueError:
            virtual_core = VirtualCoreSensor.T1

        raw_virtual_surface = (byte >> 3) & VirtualSurfaceSensor.MASK.value
        try:
            virtual_surface = VirtualSurfaceSensor(raw_virtual_surface)
        except ValueError:
            virtual_surface = VirtualSurfaceSensor.T4

        raw_virtual_ambient = (byte >> 5) & VirtualAmbientSensor.MASK.value
        try:
            virtual_ambient = VirtualAmbientSensor(raw_virtual_ambient)
        except ValueError:
            virtual_ambient = VirtualAmbientSensor.T5

        return VirtualSensors(virtual_core, virtual_surface, virtual_ambient)



from enum import Enum


class PredictionType(Enum):
    """Prediction type."""

    NONE = 0x00
    """No prediction type."""

    REMOVAL = 0x01
    """Removal."""

    RESTING = 0x02
    """Resting."""

    RESERVED = 0x03

    MASK = 0x3

    def to_string(self):
        if self == PredictionType.NONE:
            return "None"
        elif self == PredictionType.REMOVAL:
            return "Removal"
        elif self == PredictionType.RESTING:
            return "Resting"
        elif self == PredictionType.RESERVED:
            return "Reserved"

        return "Unknown"



# -*- coding: utf-8 -*-

"""Top-level package for cobustion_ble."""

from combustion_ble.ble_data import __all__ as all_ble
from combustion_ble.ble_manager import BluetoothMode
from combustion_ble.device_manager import DeviceManager
from combustion_ble.devices.probe import VirtualTemperatures
from combustion_ble.version import VERSION, VERSION_SHORT

__all__ = [
    "VERSION",
    "VERSION_SHORT",
    "BluetoothMode",
    "DeviceManager",
    "devices",
    "VirtualTemperatures",
    *all_ble,
]



"""UART"""

from .log_request import LogRequest
from .log_response import LogResponse
from .message_type import MessageType
from .read_over_temperature import (
    ReadOverTemperatureRequest,
    ReadOverTemperatureResponse,
)
from .request import Request
from .response import Response
from .response_from_data import responses_from_data
from .session_info import SessionInfoRequest, SessionInfoResponse, SessionInformation
from .set_color import SetColorRequest, SetColorResponse
from .set_id import SetIDRequest, SetIDResponse
from .set_prediction import SetPredictionRequest, SetPredictionResponse

__all__ = [
    "LogRequest",
    "LogResponse",
    "MessageType",
    "ReadOverTemperatureRequest",
    "ReadOverTemperatureResponse",
    "Request",
    "responses_from_data",
    "Response",
    "SessionInformation",
    "SessionInfoRequest",
    "SessionInfoResponse",
    "SetColorRequest",
    "SetColorResponse",
    "SetIDRequest",
    "SetIDResponse",
    "SetPredictionRequest",
    "SetPredictionResponse",
]



import struct

from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.uart.meatnet.node_request import NodeRequest


class NodeReadSessionInfoRequest(NodeRequest):
    def __init__(self, serial_number: int):
        serial_number_bytes = struct.pack("<I", serial_number)
        super().__init__(
            outgoing_payload=serial_number_bytes, message_type=NodeMessageType.SESSION_INFO
        )



from combustion_ble.ble_data.prediction_log import PredictionLog
from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures
from combustion_ble.uart.meatnet.node_response import NodeResponse


class NodeReadLogsResponse(NodeResponse):
    MINIMUM_PAYLOAD_LENGTH = 28
    HEADER_LENGTH = NodeResponse.HEADER_LENGTH
    SERIAL_RANGE = slice(HEADER_LENGTH, HEADER_LENGTH + 4)
    SEQUENCE_RANGE = slice(HEADER_LENGTH + 4, HEADER_LENGTH + 8)
    TEMPERATURE_RANGE = slice(HEADER_LENGTH + 8, HEADER_LENGTH + 21)
    PREDICTION_LOG_RANGE = slice(HEADER_LENGTH + 21, HEADER_LENGTH + 28)

    def __init__(self, data, success, request_id, response_id, payload_length):
        serial_raw = data[self.SERIAL_RANGE]
        self.probe_serial_number = int.from_bytes(serial_raw, byteorder="little")

        sequence_raw = data[self.SEQUENCE_RANGE]
        self.sequence_number = int.from_bytes(sequence_raw, byteorder="little")

        temp_data = data[self.TEMPERATURE_RANGE]
        self.temperatures = ProbeTemperatures.from_raw_data(temp_data)

        prediction_log_data = data[self.PREDICTION_LOG_RANGE]
        self.prediction_log = PredictionLog.from_raw(prediction_log_data)

        super().__init__(success, request_id, response_id, payload_length)

    @classmethod
    def from_raw(cls, data, success, request_id, response_id, payload_length):
        if payload_length < cls.MINIMUM_PAYLOAD_LENGTH:
            return None

        return cls(data, success, request_id, response_id, payload_length)



from combustion_ble.ble_data.probe_status import ProbeStatus
from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures
from combustion_ble.uart import LogResponse
from combustion_ble.uart.meatnet import NodeReadLogsResponse


class LoggedProbeDataPoint:
    _sequence_num = 0  # Class variable for generating fake data

    def __init__(
        self,
        sequence_num: int | None = None,
        temperatures: ProbeTemperatures | None = None,
        virtual_core=None,
        virtual_surface=None,
        virtual_ambient=None,
        prediction_state=None,
        prediction_mode=None,
        prediction_type=None,
        prediction_set_point_temperature=None,
        prediction_value_seconds=None,
        estimated_core_temperature=None,
    ):
        self.sequence_num = sequence_num
        self.temperatures = temperatures
        self.virtual_core = virtual_core
        self.virtual_surface = virtual_surface
        self.virtual_ambient = virtual_ambient
        self.prediction_state = prediction_state
        self.prediction_mode = prediction_mode
        self.prediction_type = prediction_type
        self.prediction_set_point_temperature = prediction_set_point_temperature
        self.prediction_value_seconds = prediction_value_seconds
        self.estimated_core_temperature = estimated_core_temperature

    @classmethod
    def from_device_status(cls, device_status: ProbeStatus):
        return cls(
            sequence_num=device_status.max_sequence_number,
            temperatures=device_status.temperatures,
            virtual_core=device_status.battery_status_virtual_sensors.virtual_sensors.virtual_core,
            virtual_surface=device_status.battery_status_virtual_sensors.virtual_sensors.virtual_surface,
            virtual_ambient=device_status.battery_status_virtual_sensors.virtual_sensors.virtual_ambient,
            prediction_state=device_status.prediction_status.prediction_state,
            prediction_mode=device_status.prediction_status.prediction_mode,
            prediction_type=device_status.prediction_status.prediction_type,
            prediction_set_point_temperature=device_status.prediction_status.prediction_set_point_temperature,
            prediction_value_seconds=device_status.prediction_status.prediction_value_seconds,
            estimated_core_temperature=device_status.prediction_status.estimated_core_temperature,
        )

    @classmethod
    def from_log_response(cls, log_response: LogResponse):
        return cls(
            sequence_num=log_response.sequence_number,
            temperatures=log_response.temperatures,
            virtual_core=log_response.prediction_log.virtual_sensors.virtual_core,
            virtual_surface=log_response.prediction_log.virtual_sensors.virtual_surface,
            virtual_ambient=log_response.prediction_log.virtual_sensors.virtual_ambient,
            prediction_state=log_response.prediction_log.prediction_state,
            prediction_mode=log_response.prediction_log.prediction_mode,
            prediction_type=log_response.prediction_log.prediction_type,
            prediction_set_point_temperature=log_response.prediction_log.prediction_set_point_temperature,
            prediction_value_seconds=log_response.prediction_log.prediction_value_seconds,
            estimated_core_temperature=log_response.prediction_log.estimated_core_temperature,
        )

    @classmethod
    def from_node_read_logs_response(cls, logs_response: NodeReadLogsResponse):
        return cls(
            sequence_num=logs_response.sequence_number,
            temperatures=logs_response.temperatures,
            virtual_core=logs_response.prediction_log.virtual_sensors.virtual_core,
            virtual_surface=logs_response.prediction_log.virtual_sensors.virtual_surface,
            virtual_ambient=logs_response.prediction_log.virtual_sensors.virtual_ambient,
            prediction_state=logs_response.prediction_log.prediction_state,
            prediction_mode=logs_response.prediction_log.prediction_mode,
            prediction_type=logs_response.prediction_log.prediction_type,
            prediction_set_point_temperature=logs_response.prediction_log.prediction_set_point_temperature,
            prediction_value_seconds=logs_response.prediction_log.prediction_value_seconds,
            estimated_core_temperature=logs_response.prediction_log.estimated_core_temperature,
        )

    def __eq__(self, other):
        if isinstance(other, LoggedProbeDataPoint):
            return self.sequence_num == other.sequence_num
        return False

    def __hash__(self):
        return hash(self.sequence_num)



class InstantReadFilter:
    DEADBAND_RANGE_IN_CELSIUS = 0.05

    def __init__(self):
        self.values: tuple[float, float] = (
            InstantReadFilter.DEADBAND_RANGE_IN_CELSIUS,
            InstantReadFilter.celsius_to_fahrenheit_absolute(
                InstantReadFilter.DEADBAND_RANGE_IN_CELSIUS
            ),
        )

    def add_reading(self, temperature_in_celsius: float):
        if temperature_in_celsius is not None:
            self.values = (
                self.calculate_filtered_temperature(
                    temperature_in_celsius, self.values[0] if self.values else None, "celsius"
                ),
                self.calculate_filtered_temperature(
                    temperature_in_celsius, self.values[1] if self.values else None, "fahrenheit"
                ),
            )
        else:
            self.values = None

    def calculate_filtered_temperature(
        self, temperature_in_celsius: float, current_display_temperature: float, units: str
    ) -> float:
        deadband_range = (
            self.DEADBAND_RANGE_IN_CELSIUS
            if units == "celsius"
            else self.celsius_to_fahrenheit_difference(self.DEADBAND_RANGE_IN_CELSIUS)
        )

        temperature = (
            temperature_in_celsius
            if units == "celsius"
            else self.celsius_to_fahrenheit_absolute(temperature_in_celsius)
        )
        display_temperature = (
            round(current_display_temperature)
            if current_display_temperature is not None
            else round(temperature)
        )

        upper_bound = display_temperature + 0.5 + deadband_range
        lower_bound = display_temperature - 0.5 - deadband_range

        if temperature > upper_bound or temperature < lower_bound:
            return round(temperature)

        return float(display_temperature)

    @staticmethod
    def celsius_to_fahrenheit_difference(celsius: float):
        return (celsius * 9.0) / 5.0

    @staticmethod
    def celsius_to_fahrenheit_absolute(celsius: float):
        return InstantReadFilter.celsius_to_fahrenheit_difference(celsius) + 32.0



from combustion_ble.ble_data.prediction_mode import PredictionMode
from combustion_ble.uart.message_type import MessageType
from combustion_ble.uart.request import Request
from combustion_ble.uart.response import Response


class SetPredictionRequest(Request):
    def __init__(self, set_point_celsius: float, prediction_mode: PredictionMode):
        raw_set_point = int(set_point_celsius / 0.1)
        raw_payload = (prediction_mode.value << 10) | (raw_set_point & 0x3FF)
        super().__init__(payload=raw_payload.to_bytes(), message_type=MessageType.SET_PREDICTION)


class SetPredictionResponse(Response):
    pass



"""Prediction Log."""

from combustion_ble.ble_data.prediction_mode import PredictionMode
from combustion_ble.ble_data.prediction_state import PredictionState
from combustion_ble.ble_data.prediction_type import PredictionType
from combustion_ble.ble_data.virtual_sensors import VirtualSensors


class PredictionLog:
    def __init__(
        self,
        virtual_sensors: VirtualSensors,
        prediction_state: PredictionState,
        prediction_mode: PredictionMode,
        prediction_type: PredictionType,
        prediction_set_point_temperature: float,
        prediction_value_seconds: int,
        estimated_core_temperature: float,
    ):
        """Initialize"""
        self.virtual_sensors = virtual_sensors
        self.prediction_state = prediction_state
        self.prediction_mode = prediction_mode
        self.prediction_type = prediction_type
        self.prediction_set_point_temperature = prediction_set_point_temperature
        self.prediction_value_seconds = prediction_value_seconds
        self.estimated_core_temperature = estimated_core_temperature

    @staticmethod
    def from_raw(data: bytes):
        """Create instance from raw data."""
        virtual_sensors = VirtualSensors.from_byte(data[0])

        raw_prediction = (data[1] & 0x07) << 1 | (data[0] & 0x80) >> 7
        prediction_state = (
            PredictionState(raw_prediction)
            if raw_prediction in PredictionState._member_map_
            else PredictionState.UNKNOWN
        )

        raw_mode = (data[1] & PredictionMode.MASK.value) >> 3
        prediction_mode = (
            PredictionMode(raw_mode)
            if raw_mode in PredictionMode._member_map_
            else PredictionMode.NONE
        )

        raw_type = (data[1] & PredictionType.MASK.value) >> 5
        prediction_type = (
            PredictionType(raw_type)
            if raw_type in PredictionType._member_map_
            else PredictionType.NONE
        )

        raw_set_point = (data[3] & 0x01) << 9 | data[2] << 1 | (data[1] & 0x80) >> 7
        prediction_set_point_temperature = float(raw_set_point) * 0.1

        prediction_value_seconds = (data[5] & 0x03) << 15 | data[4] << 7 | (data[3] & 0xFE) >> 1

        raw_core = (data[6] & 0x1F) << 6 | (data[5] & 0xFC) >> 2
        estimated_core_temperature = (float(raw_core) * 0.1) - 20.0

        return PredictionLog(
            virtual_sensors,
            prediction_state,
            prediction_mode,
            prediction_type,
            prediction_set_point_temperature,
            prediction_value_seconds,
            estimated_core_temperature,
        )



import struct

from combustion_ble.logger import LOGGER
from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.uart.meatnet.node_read_firmware_revision_response import (
    NodeReadFirmwareRevisionResponse,
)
from combustion_ble.uart.meatnet.node_read_hardware_revision_response import (
    NodeReadHardwareRevisionResponse,
)
from combustion_ble.uart.meatnet.node_read_logs_response import NodeReadLogsResponse
from combustion_ble.uart.meatnet.node_read_model_info_response import (
    NodeReadModelInfoResponse,
)
from combustion_ble.uart.meatnet.node_read_session_info_response import (
    NodeReadSessionInfoResponse,
)
from combustion_ble.uart.meatnet.node_response import NodeResponse
from combustion_ble.uart.meatnet.node_set_prediction_request import (
    NodeSetPredictionResponse,
)
from combustion_ble.utilities.crc16ccitt import crc16ccitt


def node_response_from_data(data: bytes):
    # Sync bytes
    sync_bytes = data[0:2]
    sync_string = "".join(format(x, "02x") for x in sync_bytes)
    if sync_string != "cafe":
        LOGGER.debug("NodeResponse::from_data(): Missing sync bytes in response")
        return None

    # Message type
    type_byte = data[4]

    # Verify that this is a Response by checking the response type flag
    if type_byte & NodeResponse.RESPONSE_TYPE_FLAG != NodeResponse.RESPONSE_TYPE_FLAG:
        # If that 'response type' bit isn't set, this is probably a Request.
        return None

    message_type = NodeMessageType(type_byte & ~NodeResponse.RESPONSE_TYPE_FLAG)
    if message_type is None:
        LOGGER.debug("NodeResponse::from_data(): Unknown message type in response")
        return None

    # Request ID
    request_id = struct.unpack(">I", data[5:9])[0]

    # Response ID
    response_id = struct.unpack(">I", data[9:13])[0]

    # Success/Fail
    success = bool(data[13])

    # Payload Length
    payload_length = data[14]

    # CRC
    crc = int.from_bytes(data[2:4], byteorder="little")
    crc_data = data[4 : 15 + payload_length]
    calculated_crc = crc16ccitt(crc_data)

    if crc != calculated_crc:
        LOGGER.debug("NodeResponse::from_data(): Invalid CRC")
        return None

    response_length = payload_length + NodeResponse.HEADER_LENGTH
    if len(data) < response_length:
        LOGGER.debug("Bad number of bytes")
        return None

    if message_type == NodeMessageType.LOG:
        return NodeReadLogsResponse.from_raw(
            data, success, request_id, response_id, int(payload_length)
        )
    # TODO: NodeSetIDResponse (commented out in Swift impl)
    # TODO: NodeSetColorResponse (commented out in Swift impl)
    elif message_type == NodeMessageType.SESSION_INFO:
        return NodeReadSessionInfoResponse.from_raw(
            data, success, request_id, response_id, int(payload_length)
        )
    elif message_type == NodeMessageType.SET_PREDICTION:
        return NodeSetPredictionResponse(success, request_id, response_id, int(payload_length))
    elif message_type == NodeMessageType.PROBE_FIRMWARE_REVISION:
        return NodeReadFirmwareRevisionResponse.from_raw(
            data, success, request_id, response_id, int(payload_length)
        )
    elif message_type == NodeMessageType.PROBE_HARDWARE_REVISION:
        return NodeReadHardwareRevisionResponse.from_raw(
            data, success, request_id, response_id, int(payload_length)
        )
    elif message_type == NodeMessageType.PROBE_MODEL_INFORMATION:
        return NodeReadModelInfoResponse.from_raw(
            data, success, request_id, response_id, int(payload_length)
        )
    # TODO: NodeReadOverTemperatureResponse (commented out in Swift impl)
    else:
        LOGGER.debug("Unhandled node response type: [%s]", message_type)

    return NodeResponse(success, request_id, response_id, payload_length)



import struct

from combustion_ble.uart.meatnet.node_response import NodeResponse
from combustion_ble.uart.session_info import SessionInformation


class NodeReadSessionInfoResponse(NodeResponse):
    MINIMUM_PAYLOAD_LENGTH = 10
    SERIAL_RANGE = slice(NodeResponse.HEADER_LENGTH, NodeResponse.HEADER_LENGTH + 4)
    SESSION_ID_RANGE = slice(NodeResponse.HEADER_LENGTH + 4, NodeResponse.HEADER_LENGTH + 8)
    SAMPLE_PERIOD_RANGE = slice(NodeResponse.HEADER_LENGTH + 8, NodeResponse.HEADER_LENGTH + 10)

    def __init__(self, data, success, request_id, response_id, payload_length):
        # Extracting the serial number
        serial_raw = data[self.SERIAL_RANGE]
        self.probe_serial_number = struct.unpack("<I", serial_raw)[0]

        # Extracting the session ID
        session_id_raw = data[self.SESSION_ID_RANGE]
        session_id = struct.unpack("<I", session_id_raw)[0]

        # Extracting the sample period
        sample_period_raw = data[self.SAMPLE_PERIOD_RANGE]
        sample_period = struct.unpack("<H", sample_period_raw)[0]

        self.info = SessionInformation(session_id=session_id, sample_period=sample_period)

        super().__init__(success, request_id, response_id, payload_length)

    @classmethod
    def from_raw(cls, data, success, request_id, response_id, payload_length):
        if payload_length < cls.MINIMUM_PAYLOAD_LENGTH:
            return None
        return cls(data, success, request_id, response_id, payload_length)



from combustion_ble.ble_data.prediction_log import PredictionLog
from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures
from combustion_ble.uart.response import Response


class LogResponse(Response):
    MINIMUM_PAYLOAD_LENGTH = 24
    SEQUENCE_RANGE = slice(Response.HEADER_LENGTH, Response.HEADER_LENGTH + 4)
    TEMPERATURE_RANGE = slice(Response.HEADER_LENGTH + 4, Response.HEADER_LENGTH + 17)
    PREDICTION_LOG_RANGE = slice(Response.HEADER_LENGTH + 17, Response.HEADER_LENGTH + 24)

    def __init__(self, data, success, payload_length):
        self.sequence_number = int.from_bytes(data[LogResponse.SEQUENCE_RANGE], byteorder="little")

        temp_data = data[LogResponse.TEMPERATURE_RANGE]
        self.temperatures = ProbeTemperatures.from_raw_data(temp_data)

        prediction_log_data = data[LogResponse.PREDICTION_LOG_RANGE]
        self.prediction_log = PredictionLog.from_raw(prediction_log_data)

        super().__init__(success, payload_length)

    @classmethod
    def from_raw(cls, data, success, payload_length):
        if payload_length < cls.MINIMUM_PAYLOAD_LENGTH:
            return None
        return cls(data, success, payload_length)



import random
from typing import Optional

from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.utilities.crc16ccitt import crc16ccitt


class NodeRequest:
    HEADER_LENGTH = 10

    def __init__(
        self,
        message_type: Optional[NodeMessageType] = None,
        request_id=None,
        outgoing_payload: Optional[bytes] = None,
        payload_length: Optional[int] = None,
    ):
        if not outgoing_payload:
            self.payload_length = payload_length
            self.request_id = request_id
        else:
            assert message_type
            self.data = bytearray()
            self.payload_length = len(outgoing_payload)
            self.request_id = random.randint(1, 0xFFFFFFFF)

            # Sync Bytes { 0xCA, 0xFE }
            self.data += bytearray([0xCA, 0xFE])

            # Prepare data for CRC calculation
            crc_data = bytearray()
            crc_data.append(message_type.value)
            crc_data.extend(self.request_id.to_bytes(length=4, byteorder="little"))
            crc_data.append(self.payload_length)
            crc_data.extend(outgoing_payload)

            # Calculate CRC
            crc_value = crc16ccitt(crc_data)

            self.data.extend(crc_value.to_bytes(length=2, byteorder="little"))

            # Messaeg type, payload length, payload
            self.data.extend(crc_data)



import enum
from typing import Optional

from bleak import (
    AdvertisementDataCallback,
    BleakClient,
    BleakError,
    BleakGATTCharacteristic,
    BleakScanner,
)
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from combustion_ble.ble_data.advertising_data import AdvertisingData
from combustion_ble.ble_data.probe_status import ProbeStatus
from combustion_ble.const import (
    BT_MANUFACTURER_ID,
    DEVICE_STATUS_CHARACTERISTIC,
    FW_VERSION_CHARACTERISTIC,
    HW_VERSION_CHARACTERISTIC,
    MODEL_NUMBER_CHARACTERISTIC,
    SERIAL_NUMBER_CHARACTERISTIC,
    UART_RX_CHARACTERISTIC,
    UART_TX_CHARACTERISTIC,
)
from combustion_ble.exceptions import CombustionError
from combustion_ble.logger import LOGGER
from combustion_ble.uart import Request, SessionInfoRequest
from combustion_ble.uart.meatnet import NodeRequest
from combustion_ble.utilities.asyncio_utils import ensure_future


class BleManagerDelegate:
    def did_connect_to(self, identifier: str):
        pass

    def did_fail_to_connect_to(self, identifier: str):
        pass

    def did_disconnect_from(self, identifier: str):
        pass

    def handle_bootloader_advertising(self, advertising_name: str, rssi: int, identifier: str):
        pass

    def update_device_with_advertising(
        self, advertising: AdvertisingData, is_connectable: bool, rssi: int, identifier: str
    ):
        pass

    def update_device_with_status(self, identifier: str, status: ProbeStatus):
        pass

    def handle_uart_data(self, identifier: str, data: bytes):
        pass

    def update_device_fw_version(self, identifier: str, fw_version: str):
        pass

    def update_device_hw_revision(self, identifier: str, hw_revision: str):
        pass

    def update_device_serial_number(self, identifier: str, serial_number: str):
        pass

    def update_device_model_info(self, identifier: str, model_info: str):
        pass


class BluetoothMode(enum.Enum):
    """Mode for bluetooth device discovery."""

    ACTIVE = "active"
    """Active will initialize a BleakScanner to discover devices automatically."""

    PASSIVE = "passive"
    """Passive will allow you to interface this SDK with an externally-managed BleakScanner."""


class PendingGattReads:
    """Track pending GATT Read requests."""

    def __init__(self) -> None:
        self.pending_reads: dict[str, set[int]] = {}

    def add(self, identifier: str, char: BleakGATTCharacteristic):
        if identifier not in self.pending_reads:
            self.pending_reads[identifier] = set[int]()
        reads: set[int] = self.pending_reads.get(identifier, set[int]())
        reads.add(char.handle)

    def remove(self, identifier: str, char: BleakGATTCharacteristic | None):
        if identifier not in self.pending_reads or char is None:
            return
        reads: set[int] = self.pending_reads.get(identifier, set[int]())
        reads.discard(char.handle)

    def has(self, identifier: str, char: BleakGATTCharacteristic | None) -> bool:
        if identifier not in self.pending_reads or char is None:
            return False
        reads: set[int] = self.pending_reads.get(identifier, set[int]())
        return char.handle in reads


class BleManager:
    shared: "BleManager" = None  # type: ignore

    def __init__(self):
        self.clients: dict[str, BleakClient] = {}
        self.scanner: Optional[BleakScanner] = None
        self.delegate: Optional[BleManagerDelegate] = None
        self.device_status_characteristics: dict[str, BleakGATTCharacteristic] = {}
        self.uart_characteristics: dict[str, BleakGATTCharacteristic] = {}
        self.fw_revision_characteristics: dict[str, BleakGATTCharacteristic] = {}
        self.hw_revision_characteristics: dict[str, BleakGATTCharacteristic] = {}
        self.serial_number_characteristics: dict[str, BleakGATTCharacteristic] = {}
        self.model_number_characteristics: dict[str, BleakGATTCharacteristic] = {}
        self._pending_gatt_reads = PendingGattReads()
        self._pending_connections: set[str] = set()
        self.is_stopping = False

    async def init_bluetooth(
        self, mode: BluetoothMode = BluetoothMode.ACTIVE
    ) -> None | AdvertisementDataCallback:
        """Initialize Bluetooth"""
        if mode == BluetoothMode.ACTIVE:
            await self.init_bluetooth_scanning()
        elif mode == BluetoothMode.PASSIVE:
            return self.detection_callback

        raise TypeError("Unsupported mode: %s", mode)

    async def init_bluetooth_scanning(self):
        if self.scanner:
            raise CombustionError("Bluetooth has already been initialized.")
        if self.is_stopping:
            raise CombustionError("Cannot initialize bluetooth while it is stopping.")

        LOGGER.debug("Initializing bluetooth with our own BleakScanner.")
        self.scanner = BleakScanner(detection_callback=self.detection_callback)
        await self.scanner.start()

    async def stop_bluetooth(self):
        self.is_stopping = True
        if self.scanner:
            try:
                await self.scanner.stop()
            except Exception:
                LOGGER.exception("Error stopping Bleak scanner")
            if self.clients:
                client_list = [c for c in self.clients]
                for client in client_list:
                    try:
                        if self.clients[client].is_connected:
                            await self.clients[client].disconnect()
                    except Exception:
                        LOGGER.exception("Error disconnecting client")
        self.clients = {}
        self._pending_connections = set()
        self._pending_gatt_reads = PendingGattReads()
        self.scanner = None
        self.is_stopping = False

    def detection_callback(self, device: BLEDevice, advertisement_data: AdvertisementData):
        if BT_MANUFACTURER_ID not in advertisement_data.manufacturer_data:
            return

        advertising_data = AdvertisingData.from_bleak_data(
            advertisement_data.manufacturer_data[BT_MANUFACTURER_ID]
        )
        if advertising_data and self.delegate:
            self.delegate.update_device_with_advertising(
                advertising=advertising_data,
                is_connectable=True,  # TODO: support non-connectable devices
                rssi=advertisement_data.rssi,
                identifier=device.address,
            )

    async def connect(self, identifier: str):
        if not self.delegate:
            return
        if identifier in self._pending_connections:
            LOGGER.debug("Ignoring concurrent connect request for [%s]", identifier)
            return

        if identifier in self.clients:
            LOGGER.debug("Connecting to [%s] via established client", identifier)
            client = self.clients[identifier]
        else:
            LOGGER.debug("Connecting to [%s] via new client", identifier)
            client = BleakClient(
                identifier, disconnected_callback=self.disconnected_callback(identifier)
            )

        successful = False
        self._pending_connections.add(identifier)
        try:
            await client.connect()
            LOGGER.debug("Connection to [%s] successful", identifier)
            self.clients[identifier] = client
            successful = True
        except Exception as ex:
            LOGGER.debug("Failed connecting to [%s]: %s", identifier, ex)
            self.delegate.did_fail_to_connect_to(identifier)
        finally:
            self._pending_connections.discard(identifier)

        if successful:
            self.delegate.did_connect_to(identifier)
            self.handle_discovered_services(identifier, client)

    def disconnected_callback(self, identifier: str):
        def cb(client: BleakClient):
            if self.delegate:
                self.delegate.did_disconnect_from(identifier)
            if identifier in self.clients:
                del self.clients[identifier]
            if identifier in self.uart_characteristics:
                del self.uart_characteristics[identifier]
            if identifier in self.device_status_characteristics:
                del self.device_status_characteristics[identifier]
            if identifier in self.fw_revision_characteristics:
                del self.fw_revision_characteristics[identifier]
            if identifier in self.hw_revision_characteristics:
                del self.hw_revision_characteristics[identifier]
            if identifier in self.serial_number_characteristics:
                del self.serial_number_characteristics[identifier]
            if identifier in self.model_number_characteristics:
                del self.model_number_characteristics[identifier]

        return cb

    async def disconnect(self, identifier: str):
        if identifier in self.clients:
            client = self.clients[identifier]
            if client.is_connected:
                try:
                    await client.disconnect()
                except BleakError as be:
                    LOGGER.error("Error disconnecting from [%s]: %s", identifier, be)
                # Additional cleanup handled by disconnected_callback

    async def send_request(
        self, identifier: str, request: Request | NodeRequest
    ):  # todo this does not need to be async. we can ensure_future instead
        connection_peripheral = self.get_connected_peripheral(identifier)
        if not connection_peripheral:
            return

        uart_char = self.uart_characteristics.get(identifier)
        client = self.clients.get(identifier)
        try:
            if client and client.is_connected and uart_char:
                await client.write_gatt_char(uart_char, request.data, response=False)
        except BleakError as be:
            LOGGER.error("Error sending request to [%s]: %s", identifier, be)

    async def read_firmware_revision(self, identifier: str) -> None:
        connection_peripheral = self.get_connected_peripheral(identifier)
        if connection_peripheral:
            uart_char = self.fw_revision_characteristics.get(identifier)
            client = self.clients.get(identifier)
            if self._pending_gatt_reads.has(identifier, uart_char):
                LOGGER.debug(
                    "Discarding concurent request to read_firmware_revision for [%s]", identifier
                )
                return
            try:
                if client and client.is_connected and uart_char and self.delegate:
                    self._pending_gatt_reads.add(identifier, uart_char)
                    data = await client.read_gatt_char(uart_char, use_cached=True)
                    fw_version = data.decode(encoding="utf-8")
                    self.delegate.update_device_fw_version(identifier, fw_version)
            except BleakError as be:
                LOGGER.error("Error reading firmware version from [%s]: %s", identifier, be)
            finally:
                self._pending_gatt_reads.remove(identifier, uart_char)

    async def read_hardware_revision(self, identifier: str) -> None:
        connection_peripheral = self.get_connected_peripheral(identifier)
        if connection_peripheral:
            uart_char = self.hw_revision_characteristics.get(identifier)
            client = self.clients.get(identifier)
            if self._pending_gatt_reads.has(identifier, uart_char):
                LOGGER.debug(
                    "Discarding concurent request to read_hardware_revision for [%s]", identifier
                )
                return
            try:
                if client and client.is_connected and uart_char and self.delegate:
                    self._pending_gatt_reads.add(identifier, uart_char)
                    data = await client.read_gatt_char(uart_char, use_cached=True)
                    hw_revision = data.decode(encoding="utf-8")
                    self.delegate.update_device_hw_revision(identifier, hw_revision)
            except BleakError as be:
                LOGGER.error("Error reading hardware version from [%s]: %s", identifier, be)
            finally:
                self._pending_gatt_reads.remove(identifier, uart_char)

    async def read_serial_number(self, identifier: str) -> None:
        connection_peripheral = self.get_connected_peripheral(identifier)
        if connection_peripheral:
            uart_char = self.serial_number_characteristics.get(identifier)
            client = self.clients.get(identifier)
            if self._pending_gatt_reads.has(identifier, uart_char):
                LOGGER.debug(
                    "Discarding concurent request to read_serial_number for [%s]", identifier
                )
                return
            try:
                if client and client.is_connected and uart_char and self.delegate:
                    self._pending_gatt_reads.add(identifier, uart_char)
                    data = await client.read_gatt_char(uart_char, use_cached=True)
                    serial_number = data.decode(encoding="utf-8")
                    self.delegate.update_device_serial_number(identifier, serial_number)
            except BleakError as be:
                LOGGER.error("Error reading serial number from [%s]: %s", identifier, be)
            finally:
                self._pending_gatt_reads.remove(identifier, uart_char)

    async def read_model_number(self, identifier: str) -> None:
        connection_peripheral = self.get_connected_peripheral(identifier)
        if connection_peripheral:
            uart_char = self.model_number_characteristics.get(identifier)
            client = self.clients.get(identifier)
            if self._pending_gatt_reads.has(identifier, uart_char):
                LOGGER.debug(
                    "Discarding concurent request to read_model_number for [%s]", identifier
                )
                return
            try:
                if client and client.is_connected and uart_char and self.delegate:
                    self._pending_gatt_reads.add(identifier, uart_char)
                    data = await client.read_gatt_char(uart_char, use_cached=True)
                    model_number = data.decode(encoding="utf-8")
                    self.delegate.update_device_model_info(identifier, model_number)
            except BleakError as be:
                LOGGER.error("Error reading model number from [%s]: %s", identifier, be)
            finally:
                self._pending_gatt_reads.remove(identifier, uart_char)

    def get_connected_peripheral(self, identifier: str) -> BleakClient | None:
        connected_clients = [
            self.clients[c]
            for c in self.clients
            if self.clients[c].address == identifier and self.clients[c].is_connected
        ]
        if not connected_clients:
            return None

        return connected_clients[0]

    def handle_uart_data(self, identifier: str, data: bytes):
        if self.delegate:
            self.delegate.handle_uart_data(identifier, data)

    def handle_discovered_services(self, identifier: str, client: BleakClient):
        def uart_tx_notify_callback(char: BleakGATTCharacteristic, data: bytearray):
            if char.uuid == UART_TX_CHARACTERISTIC:
                self.handle_uart_data(identifier, bytes(data))
            elif char.uuid == DEVICE_STATUS_CHARACTERISTIC:
                probe_status = ProbeStatus.from_data(data)
                if probe_status and self.delegate:
                    self.delegate.update_device_with_status(identifier, probe_status)
            else:
                LOGGER.debug("uart_tx_notify_callback ignoring unknown char [%s]", char.uuid)

        for service in client.services:
            for characteristic in service.characteristics:
                if characteristic.uuid == UART_RX_CHARACTERISTIC:
                    self.uart_characteristics[identifier] = characteristic
                elif characteristic.uuid == DEVICE_STATUS_CHARACTERISTIC:
                    self.device_status_characteristics[identifier] = characteristic
                elif (
                    characteristic.uuid == FW_VERSION_CHARACTERISTIC
                    or characteristic.uuid == HW_VERSION_CHARACTERISTIC
                    or characteristic.uuid == SERIAL_NUMBER_CHARACTERISTIC
                    or characteristic.uuid == MODEL_NUMBER_CHARACTERISTIC
                ):
                    if characteristic.uuid == FW_VERSION_CHARACTERISTIC:
                        self.fw_revision_characteristics[identifier] = characteristic
                        ensure_future(
                            self.read_firmware_revision(identifier),
                            name="ble_manager[read_firmware_revision]",
                        )
                    elif characteristic.uuid == HW_VERSION_CHARACTERISTIC:
                        self.hw_revision_characteristics[identifier] = characteristic
                        ensure_future(
                            self.read_hardware_revision(identifier),
                            name="ble_manager[read_hardware_revision]",
                        )
                    elif characteristic.uuid == SERIAL_NUMBER_CHARACTERISTIC:
                        self.serial_number_characteristics[identifier] = characteristic
                        ensure_future(
                            self.read_serial_number(identifier),
                            name="ble_manager[read_serial_number]",
                        )
                    elif characteristic.uuid == MODEL_NUMBER_CHARACTERISTIC:
                        self.model_number_characteristics[identifier] = characteristic
                        ensure_future(
                            self.read_model_number(identifier),
                            name="ble_manager[read_model_number]",
                        )
                elif characteristic.uuid == UART_TX_CHARACTERISTIC:
                    if characteristic.descriptors:
                        client = self.get_connected_peripheral(identifier)
                        if client:
                            ensure_future(
                                client.start_notify(characteristic, uart_tx_notify_callback),
                                name="ble_manager[start_notify:uart_tx]",
                            )
                            status_char = self.device_status_characteristics.get(identifier)
                            if status_char:
                                ensure_future(
                                    client.start_notify(status_char, uart_tx_notify_callback),
                                    name="ble_manager[start_notify:device_status]",
                                )
                                ensure_future(
                                    self.send_request(identifier, request=SessionInfoRequest()),
                                    name="ble_manager[send_request:session_info]",
                                )


# Instantiate the BleManager singleton
BleManager.shared = BleManager()



from combustion_ble.ble_data.battery_status_virtual_sensors import (
    BatteryStatusVirtualSensors,
)
from combustion_ble.ble_data.food_safe_data import FoodSafeData
from combustion_ble.ble_data.mode_id import ModeId
from combustion_ble.ble_data.prediction_status import PredictionStatus
from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures


class ProbeStatus:
    def __init__(
        self,
        min_sequence_number: int,
        max_sequence_number: int,
        temperatures: ProbeTemperatures,
        mode_id: ModeId,
        battery_status_virtual_sensors: BatteryStatusVirtualSensors,
        prediction_status: PredictionStatus,
        food_safe_data: FoodSafeData | None,
    ):
        self.min_sequence_number = min_sequence_number
        self.max_sequence_number = max_sequence_number
        self.temperatures = temperatures
        self.mode_id = mode_id
        self.battery_status_virtual_sensors = battery_status_virtual_sensors
        self.prediction_status = prediction_status
        self.food_safe_data = food_safe_data

    @classmethod
    def from_data(cls, data):
        if len(data) < 30:
            return None

        min_sequence_number = int.from_bytes(data[0:4], "little")
        max_sequence_number = int.from_bytes(data[4:8], "little")

        temperatures = ProbeTemperatures.from_raw_data(data[8:21])
        mode_id = ModeId.from_byte(data[21])
        battery_status_virtual_sensors = BatteryStatusVirtualSensors.from_byte(data[22])

        prediction_status_bytes = data[23:30]
        prediction_status = PredictionStatus.from_bytes(prediction_status_bytes)

        if len(data) >= 44:
            food_safe_data = FoodSafeData.from_raw(data[30:40])

        return cls(
            min_sequence_number,
            max_sequence_number,
            temperatures,
            mode_id,
            battery_status_virtual_sensors,
            prediction_status,
            food_safe_data,
        )



from datetime import datetime
from typing import Any

from combustion_ble.exceptions import DFUNotImplementedError


class DFUDeviceType:
    THERMOMETER = "thermometer"
    DISPLAY = "display"
    CHARGER = "charger"
    UNKNOWN = "unknown"


class DFUManager:
    shared = None

    class DFU:
        def __init__(self, unique_identifier: str, firmware: Any, started_at: datetime):
            self.unique_identifier = unique_identifier
            self.firmware = firmware
            self.started_at = started_at

    def __init__(self):
        self.running_dfus = {}  # Dictionary of currently active DFUs
        self.default_firmware = {}  # Default firmware for each device type
        DFUManager.shared = self

    @staticmethod
    def set_default_dfu_for_type(dfu_file, dfu_type):
        raise DFUNotImplementedError()

    def unique_identifier_from(self, advertising_name):
        raise DFUNotImplementedError()

    @staticmethod
    def bootloader_type_from(advertising_name: str):
        if Constants.THERMOMETER_DFU_NAME in advertising_name:
            return DFUDeviceType.THERMOMETER
        if Constants.DISPLAY_DFU_NAME in advertising_name:
            return DFUDeviceType.DISPLAY
        if Constants.CHARGER_DFU_NAME in advertising_name:
            return DFUDeviceType.CHARGER
        return DFUDeviceType.UNKNOWN


class Constants:
    THERMOMETER_DFU_NAME = "Thermom_DFU_"
    DISPLAY_DFU_NAME = "Display_DFU_"
    CHARGER_DFU_NAME = "Charger_DFU_"
    RETRY_TIME_DELAY = 10  # seconds



"""Prediction Status."""

from combustion_ble.ble_data.prediction_mode import PredictionMode
from combustion_ble.ble_data.prediction_state import PredictionState
from combustion_ble.ble_data.prediction_type import PredictionType


class PredictionStatus:
    """Prediction Status."""

    def __init__(
        self,
        prediction_state: PredictionState,
        prediction_mode: PredictionMode,
        prediction_type: PredictionType,
        prediction_set_point_temperature: float,
        heat_start_temperature: float,
        prediction_value_seconds: float,
        estimated_core_temperature: float,
    ):
        self.prediction_state: PredictionState = prediction_state
        """Prediction state"""

        self.prediction_mode: PredictionMode = prediction_mode
        """Prediction mode"""

        self.prediction_type: PredictionType = prediction_type
        """Prediction type"""

        self.prediction_set_point_temperature: float = prediction_set_point_temperature
        """Prediction set point temperature"""

        self.heat_start_temperature: float = heat_start_temperature
        self.prediction_value_seconds: float = prediction_value_seconds
        """Predicted seconds remaining."""

        self.estimated_core_temperature: float = estimated_core_temperature
        """Estimated core temperature."""

    def to_dict(self):
        return {
            "prediction_state": self.prediction_state.value,
            "prediction_mode": self.prediction_mode.value,
            "prediction_type": self.prediction_type.value,
            "prediction_set_point_temperature": self.prediction_set_point_temperature,
            "heat_start_temperature": self.heat_start_temperature,
            "prediction_value_seconds": self.prediction_value_seconds,
            "estimated_core_temperature": self.estimated_core_temperature,
        }

    @staticmethod
    def from_bytes(bytes):
        raw_prediction_state = bytes[0] & PredictionState.MASK.value
        prediction_state = (
            PredictionState(raw_prediction_state)
            if raw_prediction_state in PredictionState._value2member_map_
            else PredictionState.UNKNOWN
        )

        raw_prediction_mode = (bytes[0] >> 4) & PredictionMode.MASK.value
        prediction_mode = (
            PredictionMode(raw_prediction_mode)
            if raw_prediction_mode in PredictionMode._value2member_map_
            else PredictionMode.NONE
        )

        raw_prediction_type = (bytes[0] >> 6) & PredictionType.MASK.value
        prediction_type = (
            PredictionType(raw_prediction_type)
            if raw_prediction_type in PredictionType._value2member_map_
            else PredictionType.NONE
        )

        raw_set_point = (bytes[2] & 0x03) << 8 | bytes[1]
        set_point = float(raw_set_point) * 0.1

        raw_heat_start = (bytes[3] & 0x0F) << 6 | (bytes[2] & 0xFC) >> 2
        heat_start = float(raw_heat_start) * 0.1

        seconds = (bytes[5] & 0x1F) << 12 | bytes[4] << 4 | (bytes[3] & 0xF0) >> 4

        raw_core = bytes[6] << 3 | (bytes[5] & 0xE0) >> 5
        estimated_core = (float(raw_core) * 0.1) - 20.0

        return PredictionStatus(
            prediction_state,
            prediction_mode,
            prediction_type,
            set_point,
            heat_start,
            seconds,
            estimated_core,
        )



from combustion_ble.uart.meatnet.node_request import NodeRequest
from combustion_ble.uart.meatnet.node_request_from_data import node_request_from_data
from combustion_ble.uart.meatnet.node_response import NodeResponse
from combustion_ble.uart.meatnet.node_response_from_data import node_response_from_data


class NodeUARTMessage:
    @staticmethod
    def from_data(data):
        messages = []

        number_bytes_read = 0

        while number_bytes_read < len(data):
            bytes_to_decode = data[number_bytes_read:]

            response = node_response_from_data(bytes_to_decode)
            if response:
                messages.append(response)
                number_bytes_read += response.payload_length + NodeResponse.HEADER_LENGTH

            else:
                request = node_request_from_data(bytes_to_decode)
                if request and request.payload_length:
                    messages.append(request)
                    number_bytes_read += request.payload_length + NodeRequest.HEADER_LENGTH
                else:
                    # Found invalid response or request, break out of while loop
                    break

        return messages



"""Representation of the battery status & virtual sensors portion of the advertisement payload."""

from enum import Enum

from combustion_ble.ble_data.virtual_sensors import (
    VirtualAmbientSensor,
    VirtualCoreSensor,
    VirtualSensors,
    VirtualSurfaceSensor,
)


class BatteryStatus(Enum):
    """Battery status enum."""

    OK = 0x00
    LOW = 0x01

    MASK = 0x1


class BatteryStatusVirtualSensors:
    """Representation of the battery status & virtual sensors portion of the advertisement payload."""

    def __init__(self, battery_status: BatteryStatus, virtual_sensors: VirtualSensors):
        """Initialize."""
        self.battery_status = battery_status
        self.virtual_sensors = virtual_sensors

    @staticmethod
    def from_byte(byte):
        """Create instance from raw byte."""
        raw_status = byte & BatteryStatus.MASK.value
        battery = BatteryStatus(raw_status)
        virtual_sensors = VirtualSensors.from_byte(byte >> 1)
        return BatteryStatusVirtualSensors(battery, virtual_sensors)

    @staticmethod
    def default_values():
        """Generate default values."""
        return BatteryStatusVirtualSensors(
            BatteryStatus.OK,
            VirtualSensors(VirtualCoreSensor.T1, VirtualSurfaceSensor.T4, VirtualAmbientSensor.T5),
        )



from combustion_ble.ble_data.advertising_data import CombustionProductType
from combustion_ble.ble_data.hop_count import HopCount
from combustion_ble.logger import LOGGER
from combustion_ble.uart.meatnet.node_request import NodeRequest


class NodeHeartbeatRequest(NodeRequest):
    class ConnectionDetail:
        PAYLOAD_LENGTH = 13
        PROBE_SERIAL_RANGE = slice(0, 4)
        NODE_SERIAL_RANGE = slice(0, 10)
        PRODUCT_TYPE_INDEX = 10
        ATTRIBUTES_INDEX = 11
        RSSI_INDEX = 12

        def __init__(
            self, present: bool, serial_number: str, product_type: CombustionProductType, rssi: int
        ):
            self.present = present
            self.serial_number = serial_number
            self.product_type = product_type
            self.rssi = rssi

        @classmethod
        def from_raw(cls, data: bytes):
            if len(data) < cls.PAYLOAD_LENGTH:
                return cls.not_present()

            attributes = data[cls.ATTRIBUTES_INDEX]
            if (attributes & 0x01) != 0x01:
                return cls.not_present()

            product_type = CombustionProductType(data[cls.PRODUCT_TYPE_INDEX])
            serial_number = ""
            if product_type == CombustionProductType.PROBE:
                serial_raw = int.from_bytes(data[cls.PROBE_SERIAL_RANGE], byteorder="little")
                serial_number = f"{serial_raw:08X}"
            elif product_type == CombustionProductType.MEAT_NET_NODE:
                try:
                    serial_number = data[cls.NODE_SERIAL_RANGE].decode("utf-8")
                except Exception:
                    serial_number = ""

            rssi = int.from_bytes(
                data[cls.RSSI_INDEX : cls.RSSI_INDEX + 1], byteorder="big", signed=True
            )

            return cls(True, serial_number, product_type, rssi)

        @classmethod
        def not_present(cls):
            return cls(False, "", CombustionProductType.UNKNOWN, 0)

    def __init__(self, data: bytes, request_id: int, payload_length: int):
        self.request_id = request_id
        self.payload_length = payload_length

        # Extract serial number
        self.serial_number = data[self.HEADER_LENGTH : self.HEADER_LENGTH + 10].decode("utf-8")

        # Extract MAC address
        mac_raw = data[self.HEADER_LENGTH + 10 : self.HEADER_LENGTH + 16]
        self.mac_address = ":".join("{:02x}".format(byte) for byte in mac_raw).upper()

        # Extract product type, hop count, and inbound
        self.product_type = CombustionProductType(data[self.HEADER_LENGTH + 16])
        self.hop_count = HopCount.from_network_info_byte(data[self.HEADER_LENGTH + 17])
        self.inbound = data[self.HEADER_LENGTH + 18] != 0x00

        # Extract connection details
        try:
            self.connection_details = []
            for i in range(4):
                start = self.HEADER_LENGTH + 19 + (i * self.ConnectionDetail.PAYLOAD_LENGTH)
                end = start + self.ConnectionDetail.PAYLOAD_LENGTH
                self.connection_details.append(self.ConnectionDetail.from_raw(data[start:end]))
        except Exception as ex:
            LOGGER.warn("Error getting connection info:", ex)

        super().__init__(request_id=request_id, payload_length=payload_length)

    @classmethod
    def from_raw(cls, data: bytes, request_id: int, payload_length: int):
        if payload_length < 71:
            return None
        return cls(data, request_id, payload_length)



import struct

from combustion_ble.logger import LOGGER
from combustion_ble.uart.meatnet.node_heartbeat_request import NodeHeartbeatRequest
from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.uart.meatnet.node_probe_status_request import NodeProbeStatusRequest
from combustion_ble.uart.meatnet.node_request import NodeRequest
from combustion_ble.uart.meatnet.node_sync_thermometer_list_request import (
    NodeSyncThermometerListRequest,
)
from combustion_ble.utilities.crc16ccitt import crc16ccitt


def node_request_from_data(data: bytes) -> NodeRequest | None:
    if data[:2] != b"\xCA\xFE":
        LOGGER.debug("Missing sync bytes in request")
        return None

    message_type_raw = data[4]
    message_type = None
    if message_type_raw in NodeMessageType._value2member_map_:
        message_type = NodeMessageType(message_type_raw)

    if message_type is None:
        LOGGER.debug("Unknown message type in request: [%s]", message_type_raw)
        return None

    # Request ID
    request_id = struct.unpack(">I", data[5:9])[0]

    # Payload Length
    payload_length = data[9]

    # CRC Check
    crc = int.from_bytes(data[2:4], byteorder="little")
    calculated_crc = crc16ccitt(data[4 : 10 + payload_length])

    if crc != calculated_crc:
        LOGGER.debug("Invalid CRC. Expected [%s] but found [%s]", calculated_crc, crc)
        return None

    if message_type == NodeMessageType.PROBE_STATUS:
        return NodeProbeStatusRequest.from_raw(data, request_id, payload_length)
    elif message_type == NodeMessageType.HEARTBEAT:
        return NodeHeartbeatRequest.from_raw(data, request_id, payload_length)
    elif message_type == NodeMessageType.SYNC_THERMOMETER_LIST:
        return NodeSyncThermometerListRequest.from_raw(data, request_id, payload_length)
    elif (
        message_type == NodeMessageType.SESSION_INFO
        or message_type == NodeMessageType.CONNECTED
        or message_type == NodeMessageType.DISCONNECTED
    ):
        # This SDK, as of now, does not need to act on these requests.
        # This also isn't implemented upstream. This if block exists to quiet the debug logger.
        pass

    LOGGER.debug("node_request_from_data:: Unhandled node request type: [%s]", message_type.name)
    return None



from typing import Optional

from combustion_ble.uart.meatnet.node_request import NodeRequest


class Thermometer:
    """Individual Thermometer entry."""

    def __init__(self, position: int, present: bool, serial_number: int) -> None:
        self.position = position
        self.present = present
        self.serial_number = serial_number
        self.serial_number_string = f"{self.serial_number:08X}"

    def __str__(self) -> str:
        return f"Position [{self.position}]; Present [{self.present}]; Serial: [{self.serial_number_string}]"

    @classmethod
    def from_raw(cls, position: int, data: bytes):
        present = bool(data[0])
        serial_number = int.from_bytes(data[1:5], byteorder="little")

        return cls(position, present, serial_number)


class NodeSyncThermometerListRequest(NodeRequest):
    PAYLOAD_LENGTH = 26

    def __init__(self, data: Optional[bytes], request_id, payload_length):
        if not data:
            return
        mac_address_index = NodeRequest.HEADER_LENGTH

        # Extracting mac address
        mac_raw = int.from_bytes(data[mac_address_index : mac_address_index + 6])
        self.mac_address = f"{mac_raw:08X}"

        self.thermometers: list[Thermometer] = []
        for i in range(1, 5):
            index_start = mac_address_index + (6 * i)
            index_end = index_start + 5
            self.thermometers.append(Thermometer.from_raw(i, data[index_start:index_end]))

        super().__init__(request_id=request_id, payload_length=payload_length)

    @classmethod
    def from_raw(cls, data, request_id, payload_length):
        if payload_length < cls.PAYLOAD_LENGTH:
            return None
        return cls(data, request_id, payload_length)



"""Utilities for consistently working with asyncio."""

import asyncio
from typing import Any, Callable, Coroutine

from bleak import BleakError

from combustion_ble.logger import LOGGER

DoneCallback = Callable[[asyncio.Task], None]


def _done_callback(task: asyncio.Task[Any]):
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    except BleakError as be:
        # This is raised specifically by the `corebluetooth` backend.
        # It remains to be seen if this is a good idea or not. Many bleak operations are retried by us,
        # and a single failure isn't something to be concerned about in most cases.
        # TODO: handle other backend implementations.
        if be.__str__() == "disconnected":
            LOGGER.debug(
                "Exception raised by task as a result of client disconnect = %r",
                task,
                exc_info=True,
            )
        else:
            LOGGER.exception("Exception raised by task = %r", task)
    except Exception:
        LOGGER.exception("Exception raised by task = %r", task)
    if task.cancelled():
        return


def ensure_future(callable: Coroutine[Any, Any, Any], name: str = "async operation"):
    future = asyncio.ensure_future(callable)
    future.set_name(name)
    future.add_done_callback(_done_callback)



from combustion_ble.ble_data.mode_id import ProbeID
from combustion_ble.uart.message_type import MessageType
from combustion_ble.uart.request import Request
from combustion_ble.uart.response import Response


class SetIDRequest(Request):
    def __init__(self, id: ProbeID):
        payload = id.value.to_bytes(1)
        super().__init__(payload, message_type=MessageType.SET_ID)


class SetIDResponse(Response):
    pass



from enum import Enum


class NodeMessageType(Enum):
    SET_ID = 1
    SET_COLOR = 2
    SESSION_INFO = 3
    LOG = 4
    SET_PREDICTION = 5
    READ_OVER_TEMPERATURE = 6

    CONNECTED = 0x40
    DISCONNECTED = 0x41
    READ_NODE_LIST = 0x42
    READ_NETWORK_TOPOLOGY = 0x43
    READ_PROBE_LIST = 0x44
    PROBE_STATUS = 0x45
    PROBE_FIRMWARE_REVISION = 0x46
    PROBE_HARDWARE_REVISION = 0x47
    PROBE_MODEL_INFORMATION = 0x48
    HEARTBEAT = 0x49
    ASSOCIATE_NODE = 0x4A
    SYNC_THERMOMETER_LIST = 0x4B  # TODO: Implement handling for SYNC_THERMOMETER_LIST



"""Food Safe Data.

NOTE: This file does not have an upstream match
"""

from enum import Enum, unique


@unique
class FoodSafeMode(Enum):
    """Food Safe Mode."""

    SIMPLIFIED = 0x00
    INTEGRATED = 0x01


class Product(Enum):
    pass


@unique
class SimplifiedProduct(Product):
    """Selected product for food-safety using the simplified mode."""

    DEFAULT = 0x00
    ANY_POULTRY = 0x01
    BEEF_CUTS = 0x02
    PORK_CUTS = 0x03
    VEAL_CUTS = 0x04
    LAMB_CUTS = 0x05
    GROUND_MEATS = 0x06
    HAM_FRESH_SMOKED = 0x07
    HAM_COOKED = 0x08
    EGGS = 0x09
    FISH = 0xA
    LEFTOVERS = 0xB
    CASSEROLES = 0xC


@unique
class IntegratedProduct(Product):
    """Selected product for food-safety using the integrated mode."""

    DEFAULT = 0x00
    BEEF = 0x01
    BEEF_GROUND = 0x02
    CHICKEN = 0x03
    CHICKEN_GROUND = 0x04
    PORK = 0x05
    PORK_GROUND = 0x06
    HAM = 0x07
    HAM_GROUND = 0x08
    TURKEY = 0x09
    TURKEY_GROUND = 0xA
    LAMB = 0xB
    LAMB_GROUND = 0xC
    FISH = 0xD
    FISH_GROUND = 0xE
    DAIRY = 0xF
    CUSTOM = 0x3FF


@unique
class Serving(Enum):
    """Serving options for which safety calculations are available."""

    SERVED_IMMEDIATELY = 0
    COOKED_AND_CHILLED = 1


class FoodSafeData:
    def __init__(
        self,
        food_safe_mode: FoodSafeMode,
        product: Product,
        serving: Serving,
        selected_threshold_reference_temperature: float,
        z_value: float,
        reference_temperature: float,
        d_value_at_rt: float,
        target_log_reduction: float,
    ) -> None:
        self.food_safe_mode = food_safe_mode
        self.product = product
        self.serving = serving
        self.selected_threshold_reference_temperature = selected_threshold_reference_temperature
        self.z_value = z_value
        self.reference_temperature = reference_temperature
        d_value_at_rt = d_value_at_rt
        target_log_reduction = target_log_reduction

    @classmethod
    def from_raw(cls, data: bytes):
        # ____LLLL LLLLDDDD DDDDDDDD DRRRRRRR RRRRRRZZ ZZZZZZZZ ZZZTTTTT TTTTTTTT SSSPPPPP PPPPPMMM
        # 00000100 01101000 11111001 00010010 11000000 00001101 10000100 11101000 00100000 00100001

        payload = list(data)
        payload.reverse()

        mode_raw = payload[9] & 0x7
        mode = FoodSafeMode(mode_raw)

        product_raw = ((payload[8] & 0x1F) << 5) | (payload[9] >> 3)
        product: Product | None = None
        if mode == FoodSafeMode.SIMPLIFIED:
            product = SimplifiedProduct(product_raw)
        elif mode == FoodSafeMode.INTEGRATED:
            product = IntegratedProduct(product_raw)

        serving_raw = payload[8] >> 5
        serving = Serving(serving_raw)

        threshold_ref_temp_raw = ((payload[6] & 0x1F) << 8) | payload[7]
        threshold_ref_temp = float(threshold_ref_temp_raw) * 0.05

        z_value_raw = ((payload[4] & 0x03) << 11) | (payload[5] << 3) | (payload[6] >> 5)
        z_value = float(z_value_raw) * 0.05

        ref_temp_raw = ((payload[3] & 0x7F) << 6) | (payload[4] >> 2)
        ref_temp = float(ref_temp_raw) * 0.05

        d_value_raw = ((payload[1] & 0xF) << 9) | (payload[2] << 1) | (payload[3] >> 7)
        d_value = float(d_value_raw) * 0.05

        target_log_reduction_raw = ((payload[0] & 0xF) << 4) | (payload[1] >> 4)
        target_log_reduction = float(target_log_reduction_raw) * 0.1

        return cls(
            food_safe_mode=mode,
            product=product,
            serving=serving,
            selected_threshold_reference_temperature=threshold_ref_temp,
            z_value=z_value,
            reference_temperature=ref_temp,
            d_value_at_rt=d_value,
            target_log_reduction=target_log_reduction,
        )



from datetime import datetime
from typing import Callable, Generic, TypeVar

T = TypeVar("T")
UpdateListener = Callable[[T], None]

RemoveListener = Callable[[], None]


class Monitorable(Generic[T]):
    def __init__(self, initial_value: T) -> None:
        self._listeners = set[UpdateListener]()
        self._value = initial_value
        self._last_update_time = datetime.now()

    @property
    def value(self):
        return self._value

    @property
    def last_updated(self):
        return self._last_update_time

    def add_update_listener(self, listener: UpdateListener[T]) -> RemoveListener:
        self._listeners.add(listener)

        # Notify new listener of current value
        if self.value:
            listener(self.value)

        def remove():
            """Remove the listener"""
            self._listeners.remove(listener)

        return remove

    def update(self, next_value: T) -> None:
        self._value = next_value
        self._last_update_time = datetime.now()
        for listener in self._listeners:
            listener(next_value)



from typing import Union

from combustion_ble.utilities.crc16ccitt import crc16ccitt


class Request:
    HEADER_SIZE = 6

    def __init__(self, payload: Union[bytes, bytearray], message_type: int):
        self.data = bytearray()

        # Sync Bytes
        self.data.extend([0xCA, 0xFE])

        # Prepare data for CRC calculation
        crc_data = bytearray()
        crc_data.append(message_type)
        crc_data.append(len(payload))
        crc_data.extend(payload)

        # Calculate CRC
        crc_value = crc16ccitt(crc_data)

        # Append CRC to data
        self.data.extend(
            crc_value.to_bytes(length=2, byteorder="little")
        )  # struct.pack(">H", crc_value))

        # Append message type, payload length, and payload
        self.data.extend(crc_data)



"""Device Manager."""

import asyncio
from typing import Callable, Optional

from bleak import AdvertisementDataCallback

from combustion_ble.ble_data.advertising_data import (
    AdvertisingData,
    CombustionProductType,
)
from combustion_ble.ble_data.hop_count import HopCount
from combustion_ble.ble_data.probe_status import ProbeStatus
from combustion_ble.ble_manager import BleManager, BleManagerDelegate, BluetoothMode
from combustion_ble.connection_manager import ConnectionManager
from combustion_ble.devices.device import Device
from combustion_ble.devices.meat_net_node import MeatNetNode
from combustion_ble.devices.probe import Probe
from combustion_ble.exceptions import DFUNotImplementedError
from combustion_ble.logger import LOGGER
from combustion_ble.message_handlers import MessageHandlers
from combustion_ble.uart import (
    LogRequest,
    LogResponse,
    ReadOverTemperatureRequest,
    ReadOverTemperatureResponse,
    Response,
    SessionInfoRequest,
    SessionInfoResponse,
    SessionInformation,
    SetColorResponse,
    SetIDResponse,
    SetPredictionResponse,
    responses_from_data,
)
from combustion_ble.uart.meatnet import (
    NodeProbeStatusRequest,
    NodeReadFirmwareRevisionRequest,
    NodeReadFirmwareRevisionResponse,
    NodeReadHardwareRevisionRequest,
    NodeReadHardwareRevisionResponse,
    NodeReadLogsRequest,
    NodeReadLogsResponse,
    NodeReadModelInfoRequest,
    NodeReadModelInfoResponse,
    NodeReadSessionInfoRequest,
    NodeReadSessionInfoResponse,
    NodeRequest,
    NodeResponse,
    NodeSetPredictionResponse,
    NodeUARTMessage,
)

DeviceListener = Callable[[list[Device], list[Device]], None]


class DeviceManager(BleManagerDelegate):
    """
    Primary interface for Combustion BLE Devices.

    This class is designed to be a singleton.

    **Example usage:**

    .. code-block:: python
        :linenos:

        async def async_setup(self):
            self.device_manager = DeviceManager()
            self.device_manager.enable_meatnet() # Optionally enable MeatNet
            await self.device_manager.init_bluetooth()

        def do_the_work(self):
            probes = self.device_manager.get_probes()
            # Have fun with probes.

            metnet_nodes = self.device_manager.get_meatnet_nodes()
            # Have fun with meatnet nodes.

        async def async_shutdown(self):
           await self.device_manager.async_stop()
    """

    MINIMUM_PREDICTION_SETPOINT_CELSIUS = 0.0
    MAXIMUM_PREDICTION_SETPOINT_CELSIUS = 100.0
    INVALID_PROBE_SERIAL_NUMBER = 0

    shared = None

    def __init__(self):
        """Initialize.
        You may only instantiate a single instance of this class within your application.
        """
        if DeviceManager.shared:
            raise RuntimeError("An instance already exists.")
        self.devices: dict[str, Device] = {}
        self.connection_manager = ConnectionManager(self)
        self.message_handlers = MessageHandlers()
        self.device_listeners: list[DeviceListener] = []
        DeviceManager.shared = self
        BleManager.shared.delegate = self
        self.timer_task: asyncio.Task | None = asyncio.create_task(self._start_timers())

    async def init_bluetooth(
        self, mode: BluetoothMode = BluetoothMode.ACTIVE
    ) -> Optional[AdvertisementDataCallback]:
        """Initialize bluetooth operations."""
        return await BleManager.shared.init_bluetooth(mode=mode)

    def add_device_listener(self, listener: DeviceListener) -> None:
        """Add a device listener to be notified when devices are added or removed."""
        self.device_listeners.append(listener)

    def clear_device_listeners(self) -> None:
        """Remove all device listeners."""
        self.device_listeners = []

    async def async_stop(self):
        """Stop all asynchronous tasks and BLE scanning. Must be called prior to terminating your application."""
        if self.timer_task:
            self.timer_task.cancel()
            self.timer_task = None

        # Attempt to disconnect from all devices.
        for key in self.devices:
            try:
                device = self.devices[key]
                await device.disconnect()
                if isinstance(device, Probe):
                    device.stop_session_request_timer()
                    self.connection_manager.clear_handlers_for_probe(
                        device, msg="device_manager::async_stop"
                    )
            except Exception:
                LOGGER.exception(
                    "Error disconnecting client [%s] during DeviceManager shutdown.", device
                )

        try:
            await BleManager.shared.stop_bluetooth()
        except Exception:
            LOGGER.exception("Error stopping BleManager during DeviceManager shutdown.")

    async def _start_timers(self):
        while True:
            self._update_device_stale_status()
            self.message_handlers.check_for_timeout()
            await asyncio.sleep(1)

    def _update_device_stale_status(self):
        # Update the stale status of devices
        for key, device in self.devices.items():
            device._update_device_stale()

    def add_simulated_probe(self):
        # Placeholder for adding a simulated probe
        pass

    def enable_meatnet(self):
        self.connection_manager.meat_net_enabled = True

    def enable_dfu_mode(self, enable):
        raise DFUNotImplementedError()

    def _add_device(self, device: Device):
        self.devices[device.unique_identifier] = device
        for listener in self.device_listeners:
            listener([device], [])

    def _clear_device(self, device: Device):
        if device.unique_identifier in self.devices:
            del self.devices[device.unique_identifier]
            for listener in self.device_listeners:
                listener([], [device])

    def get_probes(self) -> list[Probe]:
        return [device for device in self.devices.values() if isinstance(device, Probe)]

    def get_meatnet_nodes(self):
        if self.connection_manager.meat_net_enabled:
            return [device for device in self.devices.values() if isinstance(device, MeatNetNode)]
        else:
            return []

    def get_nearest_probe(self) -> Optional[Probe]:
        """Returns the probe nearest to this device."""
        probes = self.get_probes()
        nearest = max(probes, key=lambda probe: probe.rssi, default=None)
        return nearest

    def get_devices(self) -> list[Device]:
        return list(self.devices.values())

    def get_nearest_device(self) -> Optional[Device]:
        nearest = max(self.get_devices(), key=lambda device: device.rssi, default=None)
        return nearest

    def _get_best_node_for_probe(self, serial_number: int) -> MeatNetNode | None:
        """Gets the best Node for communicating with a Probe."""
        found_node: MeatNetNode | None = None
        found_rssi: int = Device.MIN_RSSI
        meatnet_nodes = self.get_meatnet_nodes()
        for node in meatnet_nodes:
            # Check nodes to which we are connected to see if they have a route to the probe
            if node.connection_state == Device.ConnectionState.CONNECTED:
                # Choose node with the best RSSI that has a connection to probe
                if node.has_connection_to_probe(serial_number) and node.rssi > found_rssi:
                    found_node = node
                    found_rssi = node.rssi

        return found_node

    def _get_best_route_to_probe(self, serial_number) -> Device | None:
        probe = self.find_probe_by_serial_number(serial_number)
        if probe and probe.connection_state == Probe.ConnectionState.CONNECTED:
            return probe
        else:
            return self._get_best_node_for_probe(serial_number)

    def find_probe_by_serial_number(self, serial_number) -> Probe | None:
        device = self.devices.get(str(serial_number))
        if device and isinstance(device, Probe):
            return device

        return None

    async def _connect_to_device(self, device: Device):
        if device.ble_identifier:
            # If this device has a BLE identifier (advertisements are directly detected rather than through MeatNet), attempt to connect to it.
            device._update_connection_state(Device.ConnectionState.CONNECTING)
            await BleManager.shared.connect(device.ble_identifier)

    async def _disconnect_from_device(self, device: Device):
        if device.ble_identifier:
            # If this device has a BLE identifier (advertisements are directly detected rather than through MeatNet),
            # attempt to disconnect from it.
            await BleManager.shared.disconnect(device.ble_identifier)

    async def request_logs_from(self, device: Device, min_sequence: int, max_sequence: int):
        if isinstance(device, Probe):
            target_device = self._get_best_route_to_probe(device.serial_number)
            if isinstance(target_device, Probe) and target_device.ble_identifier:
                # Request logs directly from Probe
                request = LogRequest(min_sequence=min_sequence, max_sequence=max_sequence - 1)
                await BleManager.shared.send_request(target_device.ble_identifier, request)
            elif isinstance(target_device, MeatNetNode) and target_device.ble_identifier:
                # If the best route is through a Node, send it that way.
                node_request = NodeReadLogsRequest(
                    serial_number=device.serial_number,
                    min_sequence=min_sequence,
                    max_sequence=max_sequence,
                )
                await BleManager.shared.send_request(
                    identifier=target_device.ble_identifier, request=node_request
                )

    def set_probe_id(self, device, id, completion_handler):
        # TODO implement set_probe_id
        raise NotImplementedError()

    def set_probe_color(self, device, color, completion_handler):
        # TODO implement set_probe_color
        raise NotImplementedError()

    def set_removal_prediction(self, device, removal_temperature_c, completion_handler):
        # TODO implement set_removal_prediction
        raise NotImplementedError()

    def cancel_prediction(self, device, completion_handler):
        # TODO implement cancel_prediction
        raise NotImplementedError()

    async def read_session_info(self, probe: Probe):
        target_device = self._get_best_route_to_probe(probe.serial_number)
        if isinstance(target_device, Probe) and target_device.ble_identifier:
            # If the best route is directly to the Probe, send it that way.
            request = SessionInfoRequest()
            await BleManager.shared.send_request(
                identifier=target_device.ble_identifier, request=request
            )
        elif isinstance(target_device, MeatNetNode) and target_device.ble_identifier:
            node_request = NodeReadSessionInfoRequest(serial_number=probe.serial_number)
            await BleManager.shared.send_request(
                identifier=target_device.ble_identifier, request=node_request
            )

    async def read_firmware_version(self, probe: Probe):
        """Sends request to the device to read the probe firmware version."""
        target_device = self._get_best_route_to_probe(probe.serial_number)
        if isinstance(target_device, Probe) and target_device.ble_identifier:
            # If the best route is directly to the Probe, send it that way.
            await BleManager.shared.read_firmware_revision(target_device.ble_identifier)
        elif isinstance(target_device, MeatNetNode) and target_device.ble_identifier:
            # Otherwise, send via MeatNet Node
            request = NodeReadFirmwareRevisionRequest(serial_number=probe.serial_number)
            await BleManager.shared.send_request(
                identifier=target_device.ble_identifier, request=request
            )

    async def read_hardware_version(self, probe: Probe):
        if target_device := self._get_best_route_to_probe(probe.serial_number):
            if isinstance(target_device, Probe) and target_device.ble_identifier:
                # If the best route is directly to the Probe, send it that way.
                await BleManager.shared.read_hardware_revision(target_device.ble_identifier)
            elif isinstance(target_device, MeatNetNode) and target_device.ble_identifier:
                # Otherwise, send via MeatNet Node
                request = NodeReadHardwareRevisionRequest(serial_number=probe.serial_number)
                await BleManager.shared.send_request(
                    identifier=target_device.ble_identifier, request=request
                )

    async def read_model_info_for_probe(self, probe: Probe):
        if target_device := self._get_best_route_to_probe(probe.serial_number):
            if isinstance(target_device, Probe) and target_device.ble_identifier:
                # If the best route is directly to the Probe, send it that way.
                await BleManager.shared.read_model_number(target_device.ble_identifier)
            elif isinstance(target_device, MeatNetNode) and target_device.ble_identifier:
                # Otherwise, send via MeatNet Node
                request = NodeReadModelInfoRequest(serial_number=probe.serial_number)
                await BleManager.shared.send_request(
                    identifier=target_device.ble_identifier, request=request
                )

    async def read_model_info_for_node(self, node: MeatNetNode):
        await BleManager.shared.read_model_number(node.unique_identifier)

    async def read_over_temperature_flag(self, device: Device, completion_handler):
        if isinstance(device, Probe) and device.ble_identifier:
            # Store completion handler
            self.message_handlers.add_read_over_temperature_completion_handler(
                device_identifier=device.unique_identifier, completion_handler=completion_handler
            )

            # Send request to device
            request = ReadOverTemperatureRequest()
            await BleManager.shared.send_request(identifier=device.ble_identifier, request=request)

        # TODO send via node (awaiting upstream implementation)

    def restart_failed_upgrades_with(self, dfu_files):
        raise DFUNotImplementedError()

    def find_device_by_ble_identifier(self, identifier: str) -> Device | None:
        found_device: Device | None = None
        device = self.devices.get(identifier)
        if device:
            # This was a MeatNet Node as it was stored by its BLE UUID.
            found_device = device
        else:
            # Search through Devices to see if any Probes have a matching BLE identifier.
            for device in self.devices.values():
                if device.ble_identifier and device.ble_identifier == identifier:
                    found_device = device
                    break
        return found_device

    # Delegate methods
    def did_connect_to(self, identifier):
        device = self.find_device_by_ble_identifier(identifier)
        if not device:
            return
        device._update_connection_state(Device.ConnectionState.CONNECTED)

    def did_fail_to_connect_to(self, identifier):
        device = self.find_device_by_ble_identifier(identifier)
        if device:
            device._update_connection_state(Device.ConnectionState.FAILED)

    def did_disconnect_from(self, identifier: str):
        device = self.find_device_by_ble_identifier(identifier)
        if device:
            device._update_connection_state(Device.ConnectionState.DISCONNECTED)
            self.message_handlers.clear_handlers_for_device(identifier)

    def update_device_hw_revision(self, identifier: str, revision: str):
        if device := self.find_device_by_ble_identifier(identifier):
            device.hardware_revision = revision

    def update_device_fw_version(self, identifier: str, version: str):
        if device := self.find_device_by_ble_identifier(identifier):
            device.firmware_version = version

    def update_device_serial_number(self, identifier: str, serial_number: str):
        if (device := self.find_device_by_ble_identifier(identifier) is not None) and isinstance(
            device, MeatNetNode
        ):
            device.serial_number_string = serial_number

    def update_device_model_info(self, identifier: str, model_info: str):
        if device := self.find_device_by_ble_identifier(identifier):
            device.update_with_model_info(model_info)

    def update_device_with_status(self, identifier: str, status: ProbeStatus):
        probe = self.find_device_by_ble_identifier(identifier)
        if probe and isinstance(probe, Probe):
            probe._update_probe_status(status)
            self.connection_manager.received_status_for(probe, direct_connection=True)

    def update_device_with_node_status(
        self, serial_number: int, status: ProbeStatus, hop_count: HopCount
    ):
        if probe := self.find_probe_by_serial_number(serial_number):
            probe._update_probe_status(status, hop_count)
            self.connection_manager.received_status_for(probe, direct_connection=False)

    def update_device_with_advertising(
        self, advertising: AdvertisingData, is_connectable: bool, rssi: int, identifier: str
    ):
        """Determines which Device to create/update based on received AdvertisingData."""
        if advertising.type == CombustionProductType.PROBE:
            probe = self.update_probe_with_advertising(
                advertising, is_connectable, rssi, identifier
            )
            self.connection_manager.received_probe_advertising(probe)
        elif advertising.type == CombustionProductType.MEAT_NET_NODE:
            if not self.connection_manager.meat_net_enabled:
                return

            # Update node if it is in device list
            if (node := self.devices.get(identifier)) and isinstance(node, MeatNetNode):
                node.update_with_advertising(advertising, is_connectable, rssi)
            else:
                # Create node and add to device list
                meatnet_node = MeatNetNode(advertising, self, is_connectable, rssi, identifier)
                self._add_device(meatnet_node)

                # Update the probe associated with this advertising data
                probe = self.update_probe_with_advertising(
                    advertising, is_connectable=None, rssi=None, identifier=None
                )
                if probe:
                    # Add probe to meatnet node
                    meatnet_node.update_networked_probe(probe)
                    # Notify connection manager
                    self.connection_manager.received_probe_advertising_from_node(
                        probe, meatnet_node
                    )

    def update_probe_with_advertising(
        self,
        advertising: AdvertisingData,
        is_connectable: bool | None,
        rssi: int | None,
        identifier: str | None,
    ) -> Probe | None:
        """Searches for or creates a Device record for the Probe represented by specified AdvertisingData."""
        found_probe: Probe | None = None
        # If this advertising data was from a Probe, attempt to find its Device entry by its serial number.
        if advertising.serial_number != DeviceManager.INVALID_PROBE_SERIAL_NUMBER:
            unique_identifier = str(advertising.serial_number)
            if (probe := self.devices.get(unique_identifier)) and isinstance(probe, Probe):
                probe.update_with_advertising(advertising, is_connectable, rssi, identifier)
                found_probe = probe
            else:
                # If we don't yet have an entry for this Probe, create one.
                device = Probe(advertising, self, is_connectable, rssi, identifier)
                self._add_device(device)
                found_probe = device
        return found_probe

    def update_device_with_log_response(self, identifier: str, log_response: LogResponse):
        if not log_response.success:
            return
        if (probe := self.find_device_by_ble_identifier(identifier)) and isinstance(probe, Probe):
            probe._process_log_response(log_response)

    def update_device_with_session_information(
        self, identifier: str, session_information: SessionInformation
    ):
        if (probe := self.find_device_by_ble_identifier(identifier)) and isinstance(probe, Probe):
            probe._update_with_session_information(session_information)

    def handle_uart_data(self, identifier: str, data: bytes):
        """Processes data received over UART, which could be Responses and/or Requests depending on the source."""
        if device := self.find_device_by_ble_identifier(identifier):
            if isinstance(device, Probe):
                # If this was a Probe, treat all the data as responses
                responses = responses_from_data(data)
                for response in responses:
                    self.handle_probe_uart_response(identifier, response)
            elif isinstance(device, MeatNetNode):
                # If this was a Node, the data could be Responses and/or Requests
                messages = NodeUARTMessage.from_data(data)
                for message in messages:
                    if isinstance(message, NodeRequest):
                        self.handle_node_uart_request(identifier, message)
                    elif isinstance(message, NodeResponse):
                        self.handle_node_uart_response(identifier, message)

    def handle_probe_uart_response(self, identifier: str, response: Response):
        """Probe direct message handling"""
        if isinstance(response, LogResponse):
            self.update_device_with_log_response(identifier, response)
        elif isinstance(response, SetIDResponse):
            self.message_handlers.call_set_id_completion_handler(identifier, response)
        elif isinstance(response, SetColorResponse):
            self.message_handlers.call_set_color_completion_handler(identifier, response)
        elif isinstance(response, SessionInfoResponse):
            if response.success:
                self.update_device_with_session_information(identifier, response.info)
        elif isinstance(response, SetPredictionResponse):
            self.message_handlers.call_set_prediction_completion_handler(identifier, response)
        elif isinstance(response, ReadOverTemperatureResponse):
            self.message_handlers.call_read_over_temperature_completion_handler(
                identifier, response
            )

    def handle_node_uart_request(self, identifier: str, request: NodeRequest):
        if isinstance(request, NodeProbeStatusRequest):
            probe_status = request.probe_status
            hop_count = request.hop_count
            self.update_device_with_node_status(request.serial_number, probe_status, hop_count)

            # Ensure the Node that sent this item has the Probe in its list of repeated devices.
            if (node := self.find_device_by_ble_identifier(identifier)) and isinstance(
                node, MeatNetNode
            ):
                if probe := self.find_probe_by_serial_number(request.serial_number):
                    node.update_networked_probe(probe)
        # elif isinstance(request, NodeSyncThermometerListRequest):
        #     if (node := self.find_device_by_ble_identifier(identifier)) and isinstance(
        #         node, MeatNetNode
        #     ):
        #         LOGGER.info("Node [%s] is connected to probes [%s]", request.mac_address, [t.__str__() for t in request.thermometers])

        # elif isinstance(request, NodeHeartbeatRequest):
        #     # TODO handle heartbeat request (commented-out in Swift impl)
        #     return

    def handle_node_uart_response(self, identifier: str, response: NodeResponse):
        if isinstance(response, NodeSetPredictionResponse):
            self.message_handlers.call_node_set_prediction_completion_handler(identifier, response)
        elif isinstance(response, NodeReadFirmwareRevisionResponse):
            probe = self.find_probe_by_serial_number(serial_number=response.probe_serial_number)
            if probe:
                probe.firmware_version = response.fw_revision
        elif isinstance(response, NodeReadHardwareRevisionResponse):
            probe = self.find_probe_by_serial_number(serial_number=response.probe_serial_number)
            if probe:
                probe.hardware_revision = response.hw_revision
        elif isinstance(response, NodeReadModelInfoResponse):
            probe = self.find_probe_by_serial_number(serial_number=response.probe_serial_number)
            if probe:
                probe.update_with_model_info(response.model_info)
        elif isinstance(response, NodeReadSessionInfoResponse):
            probe = self.find_probe_by_serial_number(serial_number=response.probe_serial_number)
            if probe:
                probe._update_with_session_information(response.info)
        elif isinstance(response, NodeReadLogsResponse):
            probe = self.find_probe_by_serial_number(serial_number=response.probe_serial_number)
            if probe:
                probe._process_log_response(log_response=response)



import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, Optional

from combustion_ble.devices.device import Device
from combustion_ble.utilities.asyncio_utils import ensure_future

if TYPE_CHECKING:
    from combustion_ble.device_manager import DeviceManager
    from combustion_ble.devices.meat_net_node import MeatNetNode
    from combustion_ble.devices.probe import Probe


class ConnectionManager:
    def __init__(self, device_manager: "DeviceManager"):
        self.meat_net_enabled = False
        self.dfu_mode_enabled = False
        self.connection_timers: dict[str, asyncio.Task] = {}
        self.last_status_update: dict[str, datetime] = {}
        self.PROBE_STATUS_STALE_TIMEOUT = 10.0
        self.device_manager = device_manager

    def received_probe_advertising(self, probe: Optional["Probe"]):
        if probe is None:
            return

        probe_status_stale = True

        if probe.serial_number_string in self.last_status_update:
            last_update_time = self.last_status_update[probe.serial_number_string]
            probe_status_stale = (
                datetime.now() - last_update_time
            ).total_seconds() > self.PROBE_STATUS_STALE_TIMEOUT

        if self.dfu_mode_enabled:
            ensure_future(probe.connect(), "probe.connect[dfu]")
        elif (
            self.meat_net_enabled
            and probe_status_stale
            and probe.connection_state != "connected"
            and probe.serial_number_string not in self.connection_timers
        ):
            self.connection_timers[probe.serial_number_string] = asyncio.create_task(
                self.connect_probe_after_delay(probe)
            )

    def clear_handlers_for_probe(self, probe: "Probe", msg: str = "None"):
        """Clear all handlers for the given probe."""
        if probe.serial_number_string in self.connection_timers:
            self.connection_timers[probe.serial_number_string].cancel(msg)
            del self.connection_timers[probe.serial_number_string]

    async def connect_probe_after_delay(self, probe: "Probe"):
        await asyncio.sleep(3)
        updated_probe = self.get_probe_with_serial(probe.serial_number_string)
        if updated_probe:
            await updated_probe.connect()
        del self.connection_timers[probe.serial_number_string]

    def received_probe_advertising_from_node(self, probe: Optional["Probe"], node: "MeatNetNode"):
        if self.meat_net_enabled:
            ensure_future(node.connect(), "probe.connect[meatnet]")

    def received_status_for(self, probe: "Probe", direct_connection: bool):
        self.last_status_update[probe.serial_number_string] = datetime.now()

        if not direct_connection and self.meat_net_enabled and not self.dfu_mode_enabled:
            updated_probe = self.get_probe_with_serial(probe.serial_number_string)
            if updated_probe and updated_probe.connection_state == Device.ConnectionState.CONNECTED:
                ensure_future(updated_probe.disconnect(), "probe.disconnect[prefer_meatnet]")

    def get_probe_with_serial(self, serial: str) -> Optional["Probe"]:
        probes = self.device_manager.get_probes()
        return next((probe for probe in probes if probe.serial_number_string == serial), None)



import struct
from typing import Optional

from combustion_ble.ble_data.hop_count import HopCount
from combustion_ble.ble_data.probe_status import ProbeStatus
from combustion_ble.uart.meatnet.node_request import NodeRequest


class NodeProbeStatusRequest(NodeRequest):
    PAYLOAD_LENGTH = 35

    def __init__(self, data: Optional[bytes], request_id, payload_length):
        if not data:
            return
        sequence_byte_index = NodeRequest.HEADER_LENGTH

        # Extracting the serial number
        serial_number_raw = data[sequence_byte_index : sequence_byte_index + 4]
        self.serial_number = struct.unpack("<I", serial_number_raw)[0]

        # Parse Probe Status
        probe_status_raw = data[sequence_byte_index + 4 : sequence_byte_index + 48]
        self.probe_status = ProbeStatus.from_data(probe_status_raw)

        # Extracting Hop Count
        hop_count_raw = data[sequence_byte_index + 34 : sequence_byte_index + 35]
        hop_count_integer = hop_count_raw[0]
        self.hop_count = HopCount.from_network_info_byte(hop_count_integer)

        super().__init__(request_id=request_id, payload_length=payload_length)

    @classmethod
    def from_raw(cls, data, request_id, payload_length):
        if payload_length < cls.PAYLOAD_LENGTH:
            return None
        return cls(data, request_id, payload_length)



import struct

from combustion_ble.uart.meatnet.node_message_type import NodeMessageType
from combustion_ble.uart.meatnet.node_request import NodeRequest
from combustion_ble.uart.meatnet.node_response import NodeResponse


class NodeSetPredictionRequest(NodeRequest):
    def __init__(self, serial_number, set_point_celsius, mode):
        # Packing serial number into bytes
        serial_number_bytes = struct.pack("<I", serial_number)

        # Calculating raw set point and payload
        raw_set_point = int(set_point_celsius / 0.1)
        raw_payload = (mode.value << 10) | (raw_set_point & 0x3FF)
        raw_payload_bytes = struct.pack("<H", raw_payload)

        # Combining both parts of the payload
        payload = serial_number_bytes + raw_payload_bytes

        super().__init__(outgoing_payload=payload, message_type=NodeMessageType.SET_PREDICTION)


class NodeSetPredictionResponse(NodeResponse):
    pass



from combustion_ble.uart.message_type import MessageType
from combustion_ble.uart.request import Request
from combustion_ble.uart.response import Response


class ReadOverTemperatureRequest(Request):
    def __init__(self):
        super().__init__(payload=bytes(), message_type=MessageType.READ_OVER_TEMPERATURE)


class ReadOverTemperatureResponse(Response):
    PAYLOAD_LENGTH = 1

    def __init__(self, data, success, payload_length):
        sequence_byte_index = Response.HEADER_LENGTH
        flag_set_byte = data[sequence_byte_index : sequence_byte_index + 1]

        self.flag_set = bool(flag_set_byte[0])

        super().__init__(success, payload_length)



def crc16ccitt(data):
    """Calculate the CRC-16-CCITT checksum of a byte array.
    :param data: The input data as a byte array.
    :return: The calculated CRC as an integer.
    """
    poly = 0x1021
    crc = 0xFFFF

    for byte in data:
        crc ^= byte << 8
        for _ in range(8):
            crc = (crc << 1) ^ poly if (crc & 0x8000) else crc << 1
            crc &= 0xFFFF  # Truncate to 16 bits

    return crc



from datetime import datetime
from typing import TYPE_CHECKING, Optional

from combustion_ble.exceptions import DFUNotImplementedError
from combustion_ble.utilities.asyncio_utils import ensure_future
from combustion_ble.utilities.monitor import Monitorable, RemoveListener, UpdateListener

if TYPE_CHECKING:
    from combustion_ble.device_manager import DeviceManager


class Device:
    MIN_RSSI = -128
    STALE_TIMEOUT = 15.0

    class ConnectionState:
        DISCONNECTED = "disconnected"
        CONNECTING = "connecting"
        CONNECTED = "connected"
        FAILED = "failed"

    class DFUErrorMessage:
        def __init__(self, error, message):
            self.error = error
            self.message = message

    class DFUUploadProgress:
        def __init__(self, part, total_parts, progress):
            self.part = part
            self.total_parts = total_parts
            self.progress = progress

    def __init__(
        self,
        unique_identifier: str,
        device_manager: "DeviceManager",
        ble_identifier: Optional[str] = None,
        rssi=None,
    ):
        self.unique_identifier: str = unique_identifier
        self.ble_identifier: Optional[str] = ble_identifier if ble_identifier else None
        self._rssi: Monitorable[int] = Monitorable(rssi if rssi is not None else self.MIN_RSSI)
        self.firmware_version: Optional[str] = None
        self.hardware_revision: Optional[str] = None
        self.sku: Optional[str] = None
        self.manufacturing_lot: Optional[str] = None
        self.connection_state: str = Device.ConnectionState.DISCONNECTED
        self.is_connectable: bool = False
        self.maintaining_connection: bool = False
        self.stale: bool = False
        self.dfu_state = None
        self.dfu_error = None
        self.dfu_upload_progress = None
        self.last_update_time: datetime = datetime.now()
        self.dfu_service_controller = None
        self.device_manager: "DeviceManager" = device_manager

    @property
    def rssi(self) -> int:
        """The current RSSI."""
        return self._rssi.value

    def add_rssi_listener(self, listener: UpdateListener[int]) -> RemoveListener:
        """Add a listener for RSSI changes."""
        return self._rssi.add_update_listener(listener)

    def _update_connection_state(self, state: str):
        self.connection_state = state

        if self.connection_state == Device.ConnectionState.DISCONNECTED:
            self.firmware_version = None

        if self.maintaining_connection and (
            self.connection_state == Device.ConnectionState.DISCONNECTED
            or self.connection_state == Device.ConnectionState.FAILED
        ):
            ensure_future(self.connect(), name="device.connect[update_connection_state]")

    def _update_device_stale(self):
        self.stale = (datetime.now() - self.last_update_time).total_seconds() > self.STALE_TIMEOUT
        if self.stale:
            self.is_connectable = False

    def is_dfu_running(self) -> bool:
        if not self.dfu_state:
            return False

        raise DFUNotImplementedError()

    def dfu_complete(self):
        raise DFUNotImplementedError()

    def update_with_model_info(self, model_info: str):
        # Parse the SKU and lot number, which are delimited by a ':'
        parts = model_info.split(":")
        if len(parts) == 2:
            self.sku = parts[0]
            self.manufacturing_lot = parts[1]

    async def connect(self):
        self.maintaining_connection = True

        if self.connection_state != Device.ConnectionState.CONNECTED:
            await self.device_manager._connect_to_device(self)

    async def disconnect(self):
        self.maintaining_connection = False
        await self.device_manager._disconnect_from_device(self)

    def run_software_upgrade(self, dfu_file):
        raise DFUNotImplementedError()

    def dfu_state_did_change(self, state):
        raise DFUNotImplementedError()

    def dfu_error_did_occur(self, error, message):
        raise DFUNotImplementedError()

    def dfu_progress_did_change(self, part, total_parts, progress):
        raise DFUNotImplementedError()

    def log_with_level(self, level, message):
        raise DFUNotImplementedError()

    # Hashable implementation
    def __hash__(self):
        return hash(self.unique_identifier)

    def __eq__(self, other: object):
        if not isinstance(other, Device):
            return False
        return self.unique_identifier == other.unique_identifier if other else False



"""ModeId portion of advertisement payload."""

from enum import Enum, unique


@unique
class ProbeID(Enum):
    """Probe ID."""

    ID1 = 0x00
    ID2 = 0x01
    ID3 = 0x02
    ID4 = 0x03
    ID5 = 0x04
    ID6 = 0x05
    ID7 = 0x06
    ID8 = 0x07


@unique
class ProbeColor(Enum):
    """Probe Color."""

    color1 = 0x00
    color2 = 0x01
    color3 = 0x02
    color4 = 0x03
    color5 = 0x04
    color6 = 0x05
    color7 = 0x06
    color8 = 0x07


@unique
class ProbeMode(Enum):
    """Probe Mode."""

    NORMAL = 0x00
    INSTANT_READ = 0x01
    RESERVED = 0x02
    ERROR = 0x03


class ModeId:
    """ModeId portion of advertisement payload."""

    PROBE_ID_MASK = 0x7
    PROBE_ID_SHIFT = 5
    PROBE_COLOR_MASK = 0x7
    PROBE_COLOR_SHIFT = 2
    PROBE_MODE_MASK = 0x3

    def __init__(self, id: ProbeID, color: ProbeColor, mode: ProbeMode):
        """Initialize."""
        self.id = id
        self.color = color
        self.mode = mode

    @classmethod
    def from_byte(cls, byte) -> "ModeId":
        """Create instance from byte."""
        raw_probe_id = (byte >> cls.PROBE_ID_SHIFT) & cls.PROBE_ID_MASK
        id = ProbeID(raw_probe_id)

        raw_probe_color = (byte >> cls.PROBE_COLOR_SHIFT) & cls.PROBE_COLOR_MASK
        color = ProbeColor(raw_probe_color)

        raw_mode = byte & cls.PROBE_MODE_MASK
        mode = ProbeMode(raw_mode)

        return cls(id, color, mode)

    @staticmethod
    def default_values():
        """Generate default values."""
        return ModeId(ProbeID.ID1, ProbeColor.color1, ProbeMode.NORMAL)



"""Predictive Probe."""

import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, Coroutine, Optional

from combustion_ble.ble_data import AdvertisingData, CombustionProductType
from combustion_ble.ble_data.battery_status_virtual_sensors import BatteryStatus
from combustion_ble.ble_data.hop_count import HopCount
from combustion_ble.ble_data.mode_id import ProbeColor, ProbeID, ProbeMode
from combustion_ble.ble_data.probe_status import ProbeStatus
from combustion_ble.ble_data.probe_temperatures import ProbeTemperatures
from combustion_ble.ble_data.virtual_sensors import VirtualSensors
from combustion_ble.devices.device import Device
from combustion_ble.instant_read_filter import InstantReadFilter
from combustion_ble.logged_probe_data_count import LoggedProbeDataPoint
from combustion_ble.prediction.prediction_info import PredictionInfo
from combustion_ble.prediction.prediction_manager import PredictionManager
from combustion_ble.probe_temperature_log import ProbeTemperatureLog
from combustion_ble.uart import LogResponse, SessionInformation
from combustion_ble.uart.meatnet import NodeReadLogsResponse
from combustion_ble.utilities.asyncio_utils import ensure_future
from combustion_ble.utilities.monitor import Monitorable, RemoveListener, UpdateListener

if TYPE_CHECKING:
    from ..device_manager import DeviceManager


DEADBAND_RANGE_IN_CELSIUS = 0.05


class VirtualTemperatures:
    """Virtual temperature values for this Probe."""

    def __init__(
        self,
        core_temperature=DEADBAND_RANGE_IN_CELSIUS,
        surface_temperature=DEADBAND_RANGE_IN_CELSIUS,
        ambient_temperature=DEADBAND_RANGE_IN_CELSIUS,
    ):
        self.core_temperature = core_temperature
        """The Core temperature, in Celsius"""

        self.surface_temperature = surface_temperature
        """The Surface temperature, in Celsius"""

        self.ambient_temperature = ambient_temperature
        """The Ambient temperature, in Celsius"""


class Overheating:
    """Information regarding sensor overheating."""

    def __init__(self, is_overheating: bool, overheating_sensors: list[int]) -> None:
        self.is_overheating: bool = is_overheating
        """Denotes if this sensor is overheating."""

        self.overheating_sensors: list[int] = overheating_sensors
        """The list of overheating sensors."""


class Probe(Device):
    """
    Predictive Probe.
    """

    INSTANT_READ_STALE_TIMEOUT = 5.0
    #  Number of seconds to ignore other lower-priority (higher hop count) sources of information for Instant Read
    INSTANT_READ_LOCK_TIMEOUT = 1.0

    # Number of seconds to ignore other lower-priority (higher hop count) sources of information for Normal Mode
    NORMAL_MODE_LOCK_TIMEOUT = 1.0

    # Number of seconds after which status notifications should be considered stale.
    STATUS_NOTIFICATION_STALE_TIMEOUT = 16.0

    # Overheating thresholds (in degrees C) for T1 and T2
    OVERHEATING_T1_T2_THRESHOLD = 105.0
    # Overheating thresholds (in degrees C) for T3
    OVERHEATING_T3_THRESHOLD = 115.0
    # Overheating thresholds (in degrees C) for T4
    OVERHEATING_T4_THRESHOLD = 125.0
    # Overheating thresholds (in degrees C) for T5-T8
    OVERHEATING_T5_T8_THRESHOLD = 300.0

    def __init__(
        self,
        advertising: AdvertisingData,
        device_manager: "DeviceManager",
        is_connectable=None,
        rssi=None,
        identifier: str | None = None,
    ):
        super().__init__(
            unique_identifier=str(advertising.serial_number),
            ble_identifier=identifier,
            device_manager=device_manager,
            rssi=rssi,
        )
        self._serial_number = advertising.serial_number
        self._serial_number_string = f"{self._serial_number:08X}"

        self._id = advertising.mode_id.id
        self._color = advertising.mode_id.color
        self._current_temperatures: Monitorable[Optional[ProbeTemperatures]] = Monitorable(None)
        self._instant_read_celsius: Optional[float] = None
        self._instant_read_fahrenheit: Optional[float] = None
        self._instant_read_temperature: Optional[float] = None
        self._min_sequence_number: Optional[int] = None
        self._max_sequence_number: Optional[int] = None
        self._percent_of_logs_synced: Optional[int] = None
        self._battery_status = Monitorable(BatteryStatus.OK)
        self._virtual_sensors: Optional[VirtualSensors] = None
        self._prediction_info: Monitorable[Optional[PredictionInfo]] = Monitorable(None)
        self._virtual_temperatures: Monitorable[VirtualTemperatures] = Monitorable(
            VirtualTemperatures()
        )
        self._temperature_logs: list[ProbeTemperatureLog] = []
        self._overheating: Monitorable[Overheating] = Monitorable(
            Overheating(is_overheating=False, overheating_sensors=[])
        )
        self._last_status_notification_time = datetime.now()
        self._status_notifications_stale = False
        self._session_information: Optional[SessionInformation] = None
        self._last_instant_read: Optional[datetime] = None
        self._last_instant_read_hop_count: Optional[HopCount] = None
        self._last_normal_mode: Optional[datetime] = None
        self._last_normal_mode_hop_count: Optional[HopCount] = None
        self._prediction_manager = PredictionManager()
        self._instant_read_filter = InstantReadFilter()
        self._session_request_task: Optional[asyncio.Task] = None

        self._prediction_manager.add_update_listener(self._publish_prediction_info)

        # Update the probe with advertising data
        self.update_with_advertising(advertising, is_connectable, rssi, identifier)

        # Start timer to re-request session information every 3 minutes
        self.start_session_request_timer()

    def as_dict(self) -> dict:
        """Dictionary representation of this device. Required for the orjson encoder to properly encode this class."""
        return {"serial_number_string": self.serial_number_string}

    @property
    def serial_number(self) -> int:
        """Serial number for this device."""
        return self._serial_number

    @property
    def serial_number_string(self) -> str:
        """Readable (string) representation of this device's serial number."""
        return self._serial_number_string

    @property
    def batery_status(self) -> BatteryStatus:
        """The current battery status."""
        return self._battery_status.value

    def add_battery_status_listener(
        self, listener: UpdateListener[BatteryStatus]
    ) -> RemoveListener:
        """Add a listener for battery status changes."""
        return self._battery_status.add_update_listener(listener)

    @property
    def virtual_temperatures(self) -> VirtualTemperatures:
        """The current virtual temperatures."""
        return self._battery_status.value

    def add_virtual_temperatures_listener(
        self, listener: UpdateListener[VirtualTemperatures]
    ) -> RemoveListener:
        """Add a listener for virtual temperatures changes."""
        return self._virtual_temperatures.add_update_listener(listener)

    @property
    def overheating(self) -> Overheating:
        """Overheating information."""
        return self._overheating.value

    def add_overheating_listener(self, listener: UpdateListener[Overheating]) -> RemoveListener:
        """Add a listener for overheating changes."""
        return self._overheating.add_update_listener(listener)

    @property
    def current_temperatures(self) -> ProbeTemperatures | None:
        """Current temperature of each thermistor in the probe."""
        return self._current_temperatures.value

    def add_current_temperatures_listener(
        self, listener: UpdateListener[ProbeTemperatures | None]
    ) -> RemoveListener:
        """Add a listener for temperature changes."""
        return self._current_temperatures.add_update_listener(listener)

    @property
    def prediction_info(self) -> Optional[PredictionInfo]:
        """Prediction information."""
        return self._prediction_info.value

    def add_prediction_info_listener(
        self, listener: UpdateListener[Optional[PredictionInfo]]
    ) -> RemoveListener:
        """Add a listener for prediction info changes."""
        return self._prediction_info.add_update_listener(listener)

    async def _session_request_timer(self):
        while True:
            await asyncio.sleep(180)  # Wait for 180 seconds
            await self._request_session_information()

    def start_session_request_timer(self):
        if self._session_request_task is None or self._session_request_task.done():
            self._session_request_task = asyncio.create_task(self._session_request_timer())

    def stop_session_request_timer(self):
        if self._session_request_task and not self._session_request_task.done():
            self._session_request_task.cancel()

    def _publish_prediction_info(self, prediction_info: PredictionInfo):
        self._prediction_info.update(prediction_info)

    def _update_connection_state(self, state):
        if state == self.ConnectionState.DISCONNECTED:
            self._session_information = None
        super()._update_connection_state(state)

    def _update_device_stale(self):
        """Updates the device's stale status. Clears instant read temperatures if they are stale
        and updates whether status notifications are stale.
        """
        if self._last_instant_read:
            time_since_last_instant_read = (
                datetime.now() - self._last_instant_read
            ).total_seconds()
            if time_since_last_instant_read > self.INSTANT_READ_STALE_TIMEOUT:
                self._instant_read_celsius = None
                self._instant_read_fahrenheit = None
                self._instant_read_temperature = None

        self._update_status_notifications_stale()
        super()._update_device_stale()

    def update_with_advertising(
        self,
        advertising: AdvertisingData,
        is_connectable: bool | None,
        rssi: Optional[int],
        ble_identifier: str | None,
    ):
        # Update probe with advertising data
        if rssi:
            self._rssi.update(rssi)
        if is_connectable is not None:
            self.is_connectable = self.is_connectable
        if ble_identifier is not None:
            self.ble_identifier = ble_identifier

        # Only update rest of data if not connected to probe.
        # Otherwise, rely on status notifications to update data
        if self.connection_state != self.ConnectionState.CONNECTED:
            if advertising.mode_id.mode == ProbeMode.NORMAL:
                # If we should update normal mode, do so, but since this is Advertising info
                # and does not contain Prediction information, DO NOT lock it out. We want to
                # ensure the Prediction info gets updated over a Status notification if one
                # comes in.
                if self._should_update_normal_mode(advertising.hop_count):
                    # Update ID, Color, Battery status
                    self._update_id_color_battery(
                        advertising.mode_id.id,
                        advertising.mode_id.color,
                        advertising.battery_status_virtual_sensors.battery_status,
                    )

                    # Update temperatures, virtual sensors, and check for overheating
                    self._update_temperatures(
                        advertising.temperatures,
                        advertising.battery_status_virtual_sensors.virtual_sensors,
                    )

                    self.last_update_time = datetime.now()
            elif advertising.mode_id.mode == ProbeMode.INSTANT_READ:
                #  Update Instant Read temperature, providing hop count information to prioritize it.
                hop_count = None
                if advertising.type != CombustionProductType.PROBE:
                    hop_count = advertising.hop_count

                if self._update_instant_read(
                    advertising.temperatures.values[0],
                    advertising.mode_id.id,
                    advertising.mode_id.color,
                    advertising.battery_status_virtual_sensors.battery_status,
                    hop_count,
                ):
                    self.last_update_time = datetime.now()

    def _update_id_color_battery(
        self, probe_id: ProbeID, probe_color: ProbeColor, probe_battery_status: BatteryStatus
    ):
        self._id = probe_id
        self._color = probe_color
        self._battery_status.update(probe_battery_status)

    def _update_temperatures(
        self, temperatures: ProbeTemperatures, virtual_sensors: VirtualSensors
    ):
        self._current_temperatures.update(temperatures)
        self._virtual_sensors = virtual_sensors

        core = virtual_sensors.virtual_core.temperature_from(temperatures)
        surface = virtual_sensors.virtual_surface.temperature_from(temperatures)
        ambient = virtual_sensors.virtual_ambient.temperature_from(temperatures)

        self._virtual_temperatures.update(VirtualTemperatures(core, surface, ambient))

        self._check_overheating()

    def _check_overheating(self):
        if not self.current_temperatures:
            return

        any_over_temp = False
        overheating_sensor_list: list[int] = []

        # Check T1-T2
        for i in range(0, 2):
            if self.current_temperatures.values[i] >= self.OVERHEATING_T1_T2_THRESHOLD:
                any_over_temp = True
                overheating_sensor_list.append(i)

        # Check T3
        if self.current_temperatures.values[2] >= self.OVERHEATING_T3_THRESHOLD:
            any_over_temp = True
            overheating_sensor_list.append(2)

        # Check T4
        if self.current_temperatures.values[3] >= self.OVERHEATING_T4_THRESHOLD:
            any_over_temp = True
            overheating_sensor_list.append(3)

        # Check T5-T8
        for i in range(4, 8):
            if self.current_temperatures.values[i] >= self.OVERHEATING_T5_T8_THRESHOLD:
                any_over_temp = True
                overheating_sensor_list.append(i)

        if (
            self._overheating.value.is_overheating == any_over_temp
            and self._overheating.value.overheating_sensors == overheating_sensor_list
        ):
            return

        self._overheating.update(
            Overheating(is_overheating=any_over_temp, overheating_sensors=overheating_sensor_list)
        )

    def _update_probe_status(
        self, device_status: ProbeStatus, hop_count: Optional[HopCount] = None
    ):
        # Ignore status messages that have a sequence count lower than any previously received status messages
        if self._is_old_status_update(device_status):
            return
        updated = False
        if device_status.mode_id.mode == ProbeMode.NORMAL:
            if self._should_update_normal_mode(hop_count):
                self._update_id_color_battery(
                    device_status.mode_id.id,
                    device_status.mode_id.color,
                    device_status.battery_status_virtual_sensors.battery_status,
                )

                self._min_sequence_number = device_status.min_sequence_number
                self._max_sequence_number = device_status.max_sequence_number

                ensure_future(
                    self._prediction_manager.update_prediction_status(
                        device_status.prediction_status, device_status.max_sequence_number
                    ),
                    name="update_prediction_status[probe]",
                )

                self._update_temperatures(
                    device_status.temperatures,
                    device_status.battery_status_virtual_sensors.virtual_sensors,
                )

                self._add_data_to_log(LoggedProbeDataPoint.from_device_status(device_status))

                self._last_normal_mode = datetime.now()
                self._last_normal_mode_hop_count = hop_count

                updated = True
        elif device_status.mode_id.mode == ProbeMode.INSTANT_READ:
            updated = self._update_instant_read(
                device_status.temperatures.values[0],
                probe_id=device_status.mode_id.id,
                probe_color=device_status.mode_id.color,
                probe_battery_status=device_status.battery_status_virtual_sensors.battery_status,
                hop_count=hop_count,
            )
            if updated:
                self._min_sequence_number = device_status.min_sequence_number
                self._max_sequence_number = device_status.max_sequence_number

        ensure_future(self._request_missing_data(), name="request_missing_data[probe]")

        if updated:
            current = self._get_current_temperature_log()
            if current:
                self._update_log_percent()
                missing_range = current.missing_range(
                    device_status.min_sequence_number, device_status.max_sequence_number
                )
                if missing_range:
                    ensure_future(
                        self.device_manager.request_logs_from(
                            self, min_sequence=missing_range[0], max_sequence=missing_range[1]
                        ),
                        name="request_logs_from[probe]",
                    )

        self._last_status_notification_time = datetime.now()
        self._update_status_notifications_stale()
        self.last_update_time = datetime.now()

    def _update_instant_read(
        self,
        instant_read_value: float,
        probe_id: ProbeID,
        probe_color: ProbeColor,
        probe_battery_status: BatteryStatus,
        hop_count: Optional[HopCount],
    ) -> bool:
        if self._should_update_instant_read(hop_count):
            self._last_instant_read = datetime.now()
            self._last_instant_read_hop_count = hop_count
            self._instant_read_filter.add_reading(instant_read_value)
            self._instant_read_temperature = instant_read_value
            self._instant_read_celsius = self._instant_read_filter.values[0]
            self._instant_read_fahrenheit = self._instant_read_filter.values[1]

            self._update_id_color_battery(probe_id, probe_color, probe_battery_status)

            return True
        else:
            return False

    def _update_with_session_information(self, session_information: SessionInformation):
        self._session_information = session_information

    def _update_log_percent(self) -> None:
        current_log = self._get_current_temperature_log()
        max_sequence_number = self._max_sequence_number
        min_sequence_number = self._min_sequence_number
        if max_sequence_number is None or min_sequence_number is None or current_log is None:
            return

        number_logs_from_probe = current_log.logs_in_range(
            [min_sequence_number, max_sequence_number]
        )
        number_logs_on_probe = int(max_sequence_number - min_sequence_number + 1)
        if number_logs_from_probe == number_logs_on_probe:
            self._percent_of_logs_synced = 100
        else:
            self._percent_of_logs_synced = int(
                float(number_logs_from_probe) / float(number_logs_on_probe) * 100
            )

    def _is_old_status_update(self, device_status: ProbeStatus) -> bool:
        current_temp_log = self._get_current_temperature_log()
        if current_temp_log:
            max = current_temp_log.data_points[-1]
            if max.sequence_num is None:
                return False
            return device_status.max_sequence_number < max.sequence_num
        return False

    def _get_current_temperature_log(self) -> Optional[ProbeTemperatureLog]:
        if not self._session_information:
            return None
        return next(
            (
                log
                for log in self._temperature_logs
                if log.session_information.session_id == self._session_information.session_id
            ),
            None,
        )

    def _add_data_to_log(self, data_point: LoggedProbeDataPoint) -> None:
        current = self._get_current_temperature_log()
        if current:
            current.append_data_point(data_point=data_point)
        elif self._session_information:
            log = ProbeTemperatureLog(self._session_information)
            log.append_data_point(data_point=data_point)
            self._temperature_logs.append(log)

    def _process_log_response(self, log_response: LogResponse | NodeReadLogsResponse):
        # Process log response
        if isinstance(log_response, LogResponse):
            self._add_data_to_log(LoggedProbeDataPoint.from_log_response(log_response))
        elif isinstance(log_response, NodeReadLogsResponse):
            self._add_data_to_log(LoggedProbeDataPoint.from_node_read_logs_response(log_response))

    def _update_status_notifications_stale(self):
        """Updates the status of whether the status notifications are stale.
        This is based on the time elapsed since the last status notification.
        """
        time_since_last_notification = (
            datetime.now() - self._last_status_notification_time
        ).total_seconds()
        self._status_notifications_stale = (
            time_since_last_notification > self.STATUS_NOTIFICATION_STALE_TIMEOUT
        )

    async def _request_missing_data(self) -> None:
        tasks: list[Coroutine] = []
        if self._session_information is None:
            tasks.append(self.device_manager.read_session_info(self))

        if self.firmware_version is None:
            tasks.append(self.device_manager.read_firmware_version(self))

        if self.hardware_revision is None:
            tasks.append(self.device_manager.read_hardware_version(self))

        if self.manufacturing_lot is None or self.sku is None:
            tasks.append(self.device_manager.read_model_info_for_probe(self))

        await asyncio.gather(*tasks)  # TODO: error handling here, or at call site

    # Methods related to DFU functionalities
    def run_software_upgrade(self, dfu_file):
        # Placeholder for running software upgrade
        pass

    def dfu_state_did_change(self, state):
        pass

    def dfu_error_did_occur(self, error, message):
        pass

    def dfu_progress_did_change(self, part, total_parts, progress):
        pass

    def log_with_level(self, level, message):
        # Placeholder for logging
        pass

    def _should_update_normal_mode(self, hop_count: Optional[HopCount]) -> bool:
        if not hop_count or not self._last_normal_mode:
            return True
        time_since_last_normal_mode = (datetime.now() - self._last_normal_mode).total_seconds()
        if time_since_last_normal_mode > self.NORMAL_MODE_LOCK_TIMEOUT:
            return True

        if self._last_normal_mode_hop_count is None:
            return False

        if hop_count.value <= self._last_normal_mode_hop_count.value:
            return True

        return False

    def _should_update_instant_read(self, hop_count: Optional[HopCount]) -> bool:
        # If hopCount is nil, this is direct from a Probe and we should always update.
        if not hop_count or not self._last_instant_read:
            return True

        # If we haven't received Instant Read data for more than the lockout period, we should always update.
        time_since_last_instant_read = (datetime.now() - self._last_instant_read).total_seconds()
        if time_since_last_instant_read > self.INSTANT_READ_LOCK_TIMEOUT:
            return True

        # If we're in the lockout period and the last hop count was nil (i.e. direct from a Probe),
        # we should NOT update.
        if self._last_instant_read_hop_count is None:
            return False

        # Compare hop counts and see if we should update.
        if hop_count.value <= self._last_instant_read_hop_count.value:
            # This hop count is equal or better priority than the last, so update.
            return True
        else:
            # This hop is lower priority than the last, so do not update.
            return False

    async def _request_session_information(self):
        await self.device_manager.read_session_info(self)

    def __str__(self):
        return f"Probe: {self.unique_identifier}"



import asyncio
from datetime import datetime, timedelta
from typing import Optional

from combustion_ble.logged_probe_data_count import LoggedProbeDataPoint
from combustion_ble.uart import SessionInformation


class ProbeTemperatureLog:
    ACCUMULATOR_STABILIZATION_TIME = 0.2
    ACCUMULATOR_MAX = 500

    def __init__(self, session_info: SessionInformation):
        self.session_information = session_info
        self.data_points_dict: dict[int, LoggedProbeDataPoint] = {}
        self.data_point_accumulator = set[LoggedProbeDataPoint]()
        self.accumulator_timer: Optional[asyncio.Task] = None
        self.start_time: Optional[datetime] = None

    @property
    def data_points(self) -> list[LoggedProbeDataPoint]:
        """Return a list of data points, sorted by sequence number (oldest -> newest)."""
        return sorted(
            list(self.data_points_dict.values()),
            key=lambda x: x.sequence_num if x.sequence_num else -1,
        )

    def missing_range(self, sequence_range_start: int, sequence_range_end: int):
        lower_bound = None
        for search in range(sequence_range_start, sequence_range_end + 1):
            if search not in self.data_points_dict:
                lower_bound = search
                break

        if lower_bound is not None:
            upper_bound = None
            if lower_bound < sequence_range_end:
                for search in reversed(range(lower_bound + 1, sequence_range_end + 1)):
                    if search not in self.data_points_dict:
                        upper_bound = search
                        break

            if upper_bound is not None:
                return lower_bound, upper_bound
            else:
                return lower_bound, sequence_range_end

        return None

    def logs_in_range(self, sequence_numbers) -> int:
        records = 0
        if not self.data_points_dict:
            return records

        points = list(sorted(self.data_points_dict.keys()))
        min = None
        max = None
        for i in range(len(points)):
            if min is None and points[i] >= sequence_numbers[0]:
                min = i
                break

        if min is not None:
            for i in reversed(range(len(points))):
                if max is None and points[i] <= sequence_numbers[1]:
                    max = i
                    break

        if min is not None and max is not None:
            records = max - min + 1

        return records

    def insert_accumulated_data_points(self):
        added = False
        for dp in self.data_point_accumulator:
            if dp.sequence_num not in self.data_points_dict:
                assert dp.sequence_num is not None
                self.data_points_dict[dp.sequence_num] = dp
                added = True

        if added:
            self.data_points_dict = dict(sorted(self.data_points_dict.items()))

        self.data_point_accumulator.clear()

    def insert_data_point(self, new_data_point: LoggedProbeDataPoint):
        if new_data_point in self.data_point_accumulator:
            return

        self.data_point_accumulator.add(new_data_point)

        if self.accumulator_timer:
            self.accumulator_timer.cancel()

        if len(self.data_point_accumulator) > self.ACCUMULATOR_MAX:
            self.insert_accumulated_data_points()
        else:
            self.accumulator_timer = asyncio.create_task(self.accumulator_timer_task())

    async def accumulator_timer_task(self):
        await asyncio.sleep(self.ACCUMULATOR_STABILIZATION_TIME)
        await self.insert_accumulated_data_points()

    def append_data_point(self, data_point: LoggedProbeDataPoint):
        if (
            not self.data_points_dict
            or data_point.sequence_num == max(self.data_points_dict.keys()) + 1
        ):
            assert data_point.sequence_num is not None
            self.data_points_dict[data_point.sequence_num] = data_point
            if not self.start_time:
                self.set_start_time(data_point)
        else:
            self.insert_data_point(data_point)

    def set_start_time(self, data_point: LoggedProbeDataPoint):
        assert data_point.sequence_num is not None
        current_time = datetime.now()
        second_diff = (
            int(data_point.sequence_num) * int(self.session_information.sample_period) // 1000
        )
        self.start_time = current_time - timedelta(seconds=second_diff)

    @property
    def id(self):
        return self.session_information.session_id



"""Probe temperature data."""


class ProbeTemperatures:
    """Temperature values for a single probe."""

    def __init__(self, values: list[float]):
        """Initialize."""
        self.values: list[float] = values
        """Temerature readings for each of the Probe's 8 thermistors."""

    @staticmethod
    def from_reversed(bytes_: list[int]) -> "ProbeTemperatures":
        """Create instance from reversed bytes."""
        raw_temps: list[int] = []
        raw_temps.insert(0, (bytes_[0] & 0xFF) << 5 | (bytes_[1] & 0xF8) >> 3)
        raw_temps.insert(
            0, (bytes_[1] & 0x07) << 10 | (bytes_[2] & 0xFF) << 2 | (bytes_[3] & 0xC0) >> 6
        )
        raw_temps.insert(0, (bytes_[3] & 0x3F) << 7 | (bytes_[4] & 0xFE) >> 1)
        raw_temps.insert(
            0, (bytes_[4] & 0x01) << 12 | (bytes_[5] & 0xFF) << 4 | (bytes_[6] & 0xF0) >> 4
        )
        raw_temps.insert(
            0, (bytes_[6] & 0x0F) << 9 | (bytes_[7] & 0xFF) << 1 | (bytes_[8] & 0x80) >> 7
        )
        raw_temps.insert(0, (bytes_[8] & 0x7F) << 6 | (bytes_[9] & 0xFC) >> 2)
        raw_temps.insert(
            0, (bytes_[9] & 0x03) << 11 | (bytes_[10] & 0xFF) << 3 | (bytes_[11] & 0xE0) >> 5
        )
        raw_temps.insert(0, (bytes_[11] & 0x1F) << 8 | (bytes_[12] & 0xFF) >> 0)

        temperatures = [float(temp) * 0.05 - 20.0 for temp in raw_temps]
        return ProbeTemperatures(values=temperatures)

    @staticmethod
    def from_raw_data(data: bytes) -> "ProbeTemperatures":
        """Create instance from raw data."""
        bytes_ = list(data)
        bytes_.reverse()  # Reversing the byte order
        return ProbeTemperatures.from_reversed(bytes_)



"""Predication State."""

from enum import Enum


class PredictionState(Enum):
    """Prediction State."""

    PROBE_NOT_INSERTED = 0x00
    PROBE_INSERTED = 0x01
    COOKING = 0x02
    PREDICTING = 0x03
    REMOVAL_PREDICTION_DONE = 0x04
    RESERVED_STATE_5 = 0x05
    RESERVED_STATE_6 = 0x06
    RESERVED_STATE_7 = 0x07
    RESERVED_STATE_8 = 0x08
    RESERVED_STATE_9 = 0x09
    RESERVED_STATE_10 = 0x0A
    RESERVED_STATE_11 = 0x0B
    RESERVED_STATE_12 = 0x0C
    RESERVED_STATE_13 = 0x0D
    RESERVED_STATE_14 = 0x0E
    UNKNOWN = 0x0F

    MASK = 0xF

    def to_string(self):
        """Human-readable representation of prediction state."""
        if self == PredictionState.PROBE_NOT_INSERTED:
            return "Probe Not Inserted"
        elif self == PredictionState.PROBE_INSERTED:
            return "Probe Inserted"
        elif self == PredictionState.COOKING:
            return "Cooking"
        elif self == PredictionState.PREDICTING:
            return "Predicting"
        elif self == PredictionState.REMOVAL_PREDICTION_DONE:
            return "Removal Prediction Done"
        elif (
            self.value >= PredictionState.RESERVED_STATE_5.value
            and self.value <= PredictionState.RESERVED_STATE_14.value
        ):
            return f"Reserved State {self.value - 4}"
        elif self == PredictionState.UNKNOWN:
            return "Unknown"

        return "Unknown"



"""Prediction Info."""

from typing import Optional

from combustion_ble.ble_data.prediction_mode import PredictionMode
from combustion_ble.ble_data.prediction_state import PredictionState
from combustion_ble.ble_data.prediction_type import PredictionType


class PredictionInfo:
    """Prediction Info."""

    def __init__(
        self,
        prediction_state: PredictionState,
        prediction_mode: PredictionMode,
        prediction_type: PredictionType,
        prediction_set_point_temperature: float,
        estimated_core_temperature: float,
        seconds_remaining: Optional[int] = None,
        percent_through_cook: int = 0,
    ):
        """Initialize."""
        self.prediction_state = prediction_state
        self.prediction_mode = prediction_mode
        self.prediction_type = prediction_type
        self.prediction_set_point_temperature = prediction_set_point_temperature
        self.estimated_core_temperature = estimated_core_temperature
        self.seconds_remaining = seconds_remaining
        self.percent_through_cook = percent_through_cook

    def __str__(self) -> str:
        return f"Mode[{self.prediction_mode.to_string()}] Type[{self.prediction_type.to_string()}] Set Point [{round(self.prediction_set_point_temperature, 1)}] Percent Complete [{self.percent_through_cook}]"



"""MeatNet UART"""

from .node_heartbeat_request import NodeHeartbeatRequest
from .node_message_type import NodeMessageType
from .node_probe_status_request import NodeProbeStatusRequest
from .node_read_firmware_revision_request import NodeReadFirmwareRevisionRequest
from .node_read_firmware_revision_response import NodeReadFirmwareRevisionResponse
from .node_read_hardware_revision_request import NodeReadHardwareRevisionRequest
from .node_read_hardware_revision_response import NodeReadHardwareRevisionResponse
from .node_read_logs_request import NodeReadLogsRequest
from .node_read_logs_response import NodeReadLogsResponse
from .node_read_model_info_request import NodeReadModelInfoRequest
from .node_read_model_info_response import NodeReadModelInfoResponse
from .node_read_session_info_request import NodeReadSessionInfoRequest
from .node_read_session_info_response import NodeReadSessionInfoResponse
from .node_request import NodeRequest
from .node_request_from_data import node_request_from_data
from .node_response import NodeResponse
from .node_response_from_data import node_response_from_data
from .node_set_prediction_request import (
    NodeSetPredictionRequest,
    NodeSetPredictionResponse,
)
from .node_sync_thermometer_list_request import NodeSyncThermometerListRequest
from .node_uart_message import NodeUARTMessage

__all__ = [
    "NodeHeartbeatRequest",
    "NodeMessageType",
    "NodeProbeStatusRequest",
    "NodeReadFirmwareRevisionRequest",
    "NodeReadFirmwareRevisionResponse",
    "NodeReadHardwareRevisionRequest",
    "NodeReadHardwareRevisionResponse",
    "NodeReadLogsRequest",
    "NodeReadLogsResponse",
    "NodeReadModelInfoRequest",
    "NodeReadModelInfoResponse",
    "NodeReadSessionInfoRequest",
    "NodeReadSessionInfoResponse",
    "node_request_from_data",
    "NodeRequest",
    "node_response_from_data",
    "NodeResponse",
    "NodeSetPredictionRequest",
    "NodeSetPredictionResponse",
    "NodeSyncThermometerListRequest",
    "NodeUARTMessage",
]
